"""
PMAC Interface to Hexapod Hardware Controller from Symétrie.

Some of the code below is based on the dls_pmaclib Python library that was made available
as open source by Diamond Control under the LGPL v3.x. For more info and the original
source code, please go to http://controls.diamond.ac.uk/downloads/python/index.php].

Author: Rik Huygen
"""

import socket
import struct
import threading
from datetime import datetime
from datetime import timedelta
from typing import List

from egse.hexapod.symetrie import logger
from egse.hexapod.symetrie.pmac_regex import match_float_response
from egse.hexapod.symetrie.pmac_regex import match_int_response
from egse.hexapod.symetrie.pmac_regex import match_string_response

# Command set Request

VR_PMAC_SENDLINE = 0xB0
VR_PMAC_GETLINE = 0xB1
VR_PMAC_FLUSH = 0xB3
VR_PMAC_GETMEM = 0xB4
VR_PMAC_SETMEM = 0xB5
VR_PMAC_SETBIT = 0xBA
VR_PMAC_SETBITS = 0xBB
VR_PMAC_PORT = 0xBE
VR_PMAC_GETRESPONSE = 0xBF
VR_PMAC_READREADY = 0xC2
VR_CTRL_RESPONSE = 0xC4
VR_PMAC_GETBUFFER = 0xC5
VR_PMAC_WRITEBUFFER = 0xC6
VR_PMAC_WRITEERROR = 0xC7
VR_FWDOWNLOAD = 0xCB
VR_IPADDRESS = 0xE0

# Request Types

VR_DOWNLOAD = 0x40  # Command send to the device
VR_UPLOAD = 0xC0  # Command send to host

# API Function Error Codes

ERR_DISCONNECTED = -1
ERR_NOT_READY_2_BE_READ = -2
ERR_NO_CMD_MATCHING = -3
ERR_WRITE = -4
ERR_READ = -5
ERR_CONNECTION = -6
ERR_CMD_TIMEOUT = -7

# API Returns

RETURN_SUCCESS = 0
RETURN_IGNORED = -1
RETURN_FAILURE = -2

# The States of Q20 which are checked after the command has executed.

Q20_0: List[int] = []  # no checking of Q20 to be done
Q20_1: List[int] = [RETURN_SUCCESS]  # wait until success
Q20_2: List[int] = [RETURN_SUCCESS, RETURN_IGNORED]
Q20_3: List[int] = [RETURN_SUCCESS, RETURN_IGNORED, RETURN_FAILURE]

# Each command is defined as a dictionary with the following structure:
#
# {
#     'name'    : [string]   Human readable name of the command
#     'cmd'     : [string]   Actual command string for the controller.
#                            Can contain format replacement fields for commands that take input values.
#     'in'      : [int]      Number of input values. This should match up with the number of replacement fields in 'cmd'
#     'out'     : [type]     Number of output variable. The type of the output variables is defined by 'out_type'
#     'return'  : [string]   The command to retreive and check the output.
#     'check'   : [iterable] Iterable containing the output values that are expected.
#     'out_type': [function] A function object that is used to convert the output variables from string to <type func>.
# }
#
#

CMD_STOP = {
    "name": "STOP",
    "cmd": "&2 Q20=2",
    "in": 0,
    "out": 0,
    "return": "&2 Q20",
    "check": Q20_1,
    "out_type": None,
}
CMD_HOMING = {
    "name": "HOMING",
    "cmd": "&2 Q20=1",
    "in": 0,
    "out": 0,
    "return": "&2 Q20",
    "check": Q20_1,
    "out_type": None,
}
CMD_VIRTUAL_HOMING = {
    "name": "HOMINGVIRTUAL",
    "cmd": "&2 Q71={tx} Q72={ty} Q73={tz} Q74={rx} Q75={ry} Q76={rz} Q20=42",
    "in": 6,
    "out": 0,
    "return": "&2 Q20",
    "check": Q20_2,
    "out_type": None,
}
CMD_CONTROLON = {
    "name": "CONTROL_ON",
    "cmd": "&2 Q20=3",
    "in": 0,
    "out": 0,
    "return": "&2 Q20",
    "check": Q20_3,
    "out_type": None,
}
CMD_CONTROLOFF = {
    "name": "CONTROL_OFF",
    "cmd": "&2 Q20=4",
    "in": 0,
    "out": 0,
    "return": "&2 Q20",
    "check": Q20_1,
    "out_type": None,
}
CMD_MOVE = {
    "name": "MOVE",
    "cmd": "&2 Q70={cm} Q71={tx:.6f} Q72={ty:.6f} Q73={tz:.6f} Q74={rx:.6f} Q75={ry:.6f} Q76={rz:.6f} Q20=11",
    "in": 7,
    "out": 0,
    "return": "&2 Q20",
    "check": Q20_3,
    "out_type": None,
}
CMD_MAINTENANCE = {
    "name": "MAINTENANCE",
    "cmd": "&2 Q80={axis} Q20=12",
    "in": 1,
    "out": 0,
    "return": "&2 Q20",
    "check": Q20_2,
    "out_type": None,
}
CMD_SPECIFICPOS = {
    "name": "SPECIFICPOS",
    "cmd": "&2 Q80={pos} Q20=13",
    "in": 1,
    "out": 0,
    "return": "&2 Q20",
    "check": Q20_3,
    "out_type": None,
}
CMD_CLEARERROR = {
    "name": "CLEARERROR",
    "cmd": "&2 Q20=15",
    "in": 0,
    "out": 0,
    "return": "&2 Q20",
    "check": Q20_1,
    "out_type": None,
}
CMD_RESETSOFT = {
    "name": "RESETSOFT",
    "cmd": "$$$",
    "in": 0,
    "out": 0,
    "return": None,
    "check": None,
    "out_type": None,
}
CMD_JOG = {
    "name": "JOG",
    "cmd": "&2 Q68={axis} Q69={inc} Q20=41",
    "in": 0,
    "out": 0,
    "return": "&2 Q20",
    "check": Q20_2,
    "out_type": None,
}

CMD_CFG_CS = {
    "name": "CFG#CS",
    "cmd": (
        "&2 Q80={tx_u} Q81={ty_u} Q82={tz_u} Q83={rx_u} Q84={ry_u} Q85={rz_u}"
        " Q86={tx_o} Q87={ty_o} Q88={tz_o} Q89={rx_o} Q90={ry_o} Q91={rz_o} Q20=21"
    ),
    "in": 12,
    "out": 0,
    "return": "&2 Q20",
    "check": Q20_2,
    "out_type": None,
}
CMD_CFG_SPEED = {
    "name": "CFG#SPEED",
    "cmd": "&2 Q80={vt} Q81={vr} Q20=25",
    "in": 2,
    "out": 0,
    "return": "&2 Q20",
    "check": Q20_2,
    "out_type": None,
}

CMD_Q20_GET = {
    "name": "Q20?",
    "cmd": "&2 Q20",
    "in": 0,
    "out": 0,
    "return": None,
    "check": None,
    "out_type": None,
}
CMD_Q29_GET = {
    "name": "Q29?",
    "cmd": "&2 Q29",
    "in": 0,
    "out": 0,
    "return": None,
    "check": None,
    "out_type": None,
}
CMD_STATE_HEXA_GET = {
    "name": "STATE#HEXA?",
    "cmd": "&2 Q36",
    "in": 0,
    "out": 0,
    "return": None,
    "check": None,
    "out_type": None,
}
CMD_STATE_ERROR_GET = {
    "name": "STATE#ERROR?",
    "cmd": "&2 Q38",
    "in": 0,
    "out": 0,
    "return": None,
    "check": None,
    "out_type": None,
}
CMD_STATE_DEBUG_GET = {
    "name": "STATE#DEBUG?",
    "cmd": "&2 Q26,3,1 Q20",
    "in": 0,
    "out": 4,
    "return": None,
    "check": None,
    "out_type": int,
}
CMD_VERSION_GET = {
    "name": "VERSION?",
    "cmd": "&2 Q20=55",
    "in": 0,
    "out": 4,
    "return": "&2 Q20 Q80,4,1",
    "check": Q20_1,
    "out_type": bytes.decode,
}
CMD_POSUSER_GET = {
    "name": "POSUSER?",
    "cmd": "&2 Q53,6,1",
    "in": 0,
    "out": 6,
    "return": None,
    "check": None,
    "out_type": float,
}
CMD_POSVALID_GET = {
    "name": "POSVALID?",
    "cmd": "&2 Q70={cm} Q71={tx:.6f} Q72={ty:.6f} Q73={tz:.6f} Q74={rx:.6f} Q75={ry:.6f} Q76={rz:.6f} Q20=10",
    "in": 7,
    "out": 1,
    "return": "&2 Q20 Q29",
    "check": Q20_1,
    "out_type": int,
}
CMD_CFG_CS_GET = {
    "name": "CFG#CS?",
    "cmd": "&2 Q20=31",
    "in": 0,
    "out": 12,
    "return": "&2 Q20 Q80,12,1",
    "check": Q20_1,
    "out_type": float,
}
CMD_CFG_SPEED_GET = {
    "name": "CFG#SPEED?",
    "cmd": "&2 Q20=35",
    "in": 0,
    "out": 6,
    "return": "&2 Q20 Q80,6,1",
    "check": Q20_1,
    "out_type": float,
}

CMD_STRING_CID = "CID"  # Card ID number
CMD_STRING_VER = "VERSION"  # Firmware Version (Revision Level)
CMD_STRING_CPU = "CPU"
CMD_STRING_TYPE = "TYPE"
CMD_STRING_VID = "VID"  # Vendor ID
CMD_STRING_DATE = "DATE"
CMD_STRING_TIME = "TIME"
CMD_STRING_TODAY = "TODAY"

DEF_TPMAC_CMD_TIMEOUT = timedelta(seconds=5)
DEF_TPMAC_READ_TIMEOUT = timedelta(seconds=2)
DEF_TPMAC_READ_FREQ = timedelta(milliseconds=100)
DEF_TPMAC_STRING_SIZE = 1400

DEF_STRING_SIZE = 256


# #### Pure Python API interface #######################################################################################


class PMACError(Exception):
    pass


# #### Helper functions and classes ####################################################################################


def extractOutput(sourceStr, nr, func=bytes.decode):
    """
    Extract values from the source string argument.

    The output variables all have the same type and are converted using
    the func argument, e.g. int or float.

    :return: an array with the output variables.
    """
    # Remove the acknowledgement from the sourceStr

    sourceStr = sourceStr.rstrip(b"\r\x06")

    # Split the string in it's parts separated by \r

    parts = sourceStr.split(b"\r")

    # extract the values from the part and put them in out

    try:
        out = [func(part) for part in parts]
    except (ValueError, TypeError) as exc:
        raise PMACError(
            f'extractOutput(): Could not parse individual parts of "{sourceStr}" with {func}.'
        ) from exc

    return out


def extractQ20AndOutput(sourceStr, nr, func=bytes.decode):
    """
    Extract values from the source string argument.

    The first value is always the Q20 variable and is of integer type.

    The other output variables all have the same type and are converted using
    the func argument, e.g. int or float.

    Returns a tuple containing the Q20 and an array with the output variables.
    """
    # Remove the acknowledgement from the sourceStr

    sourceStr = sourceStr.rstrip(b"\r\x06")

    # Split the string in it's parts separated by \r

    parts = sourceStr.split(b"\r")

    # First value is always Q20, then the other output values

    try:
        Q20 = int(parts[0])
        out = [func(part) for part in parts[1:]]
    except (ValueError, TypeError) as ex:
        raise PMACError(
            f'extractQ20AndOutput(): Could not parse individual parts of "{sourceStr}" with {func}.'
        ) from ex

    return (Q20, out)


def decode_Q29(value):
    """
    Decode the bitfield variable Q29.

    Each bit in this variable represents a particular error in the validation of a movement.
    Several errors can be combined into the Q29 variable.

    Returns a dictionary with the bit numbers that were (on) and the corresponding error description.
    """

    # This description is taken from the API PUNA Hexapod Controller Manual, version A, section 4.5.4 POSVALID?

    description = [
        "Homing is not done",
        "Coordinate systems definition not realized",
        "Kinematic error",
        "Out of SYMETRIE (factory) workspace",
        "Out of machine workspace",
        "Out of user workspace",
        "Actuator out of limits",
        "Joints out of limits",
        "Out of limits due to backlash compensation",
        '"Abort" input enabled',
        '"Safety Sensor" inputs enabled',
    ]

    error_msgs = {}

    for bit in range(24):
        if value >> bit & 0x01:
            error_msgs[bit] = description[bit] if bit < 11 else "Reserved"

    return error_msgs


def decode_Q30(value):
    """
    Decode the bitfield variable Q30.

    Each bit in this variable represents a particular system setting of the actuator.

    Returns a dictionary with the bit numbers that were (on) and the corresponding error description.
    """

    # This description is taken from the API PUNA Hexapod Controller Manual, version A, section 4.5.5 STATE#ACTUATOR?

    description = [
        "In position",
        "Control loop on servo motors active",
        "Homing done",
        'Input "Home switch"',
        'Input "Positive limit switch"',
        'Input "Negative limit switch"',
        "Brake control output",
        "Following error (warning)",
        "Following error",
        "Actuator out of bounds error",
        "Amplifier error",
        "Encoder error",
        "Phasing error (brushless engine only)",
    ]

    error_msgs = {
        bit: description[bit] if bit < 13 else "Reserved"
        for bit in range(24)
        if value >> bit & 0x01
    }

    states = [int(x) for x in f"{value:024b}"[::-1]]

    return error_msgs, states


def decode_Q36(value):
    """
    Decode the bitfield variable Q36.

    Each bit in this variable represents the status of a particular system setting of the hexapod.

    Returns an array with the value (True/False) of the individual bits.
    """

    # This description is taken from the API PUNA Hexapod Controller Manual, version A, section 4.5.6 STATE#HEXA?

    description = [  # noqa: F841
        "Error (OR)",
        "System Initialized",
        "In position",
        "Control loop on servo motors active",
        "Homing done",
        "Brake control output",
        "Emergency stop button engaged",
        "Following error (warning)",
        "Following error",
        "Actuator out of bounds error",
        "Amplifier Error",
        "Encoder error",
        "Phasing error (brushless motors only)",
        "Homing error",
        "Kinematic error",
        '"Abort" input error',
        "R/W flash memory error",
        "Temperature error on one or several motors",
        "Home done (virtual)",
        "Encoders power off",
        "Limit switches power off",
        "Reserved",
        "Reserved",
        "Reserved",
    ]

    bit_values = [False for x in range(24)]

    for bit in range(24):
        bit_values[bit] = True if value >> bit & 0x01 else False

    return bit_values


class EthernetCommand:
    def __init__(self, requestType, request, value, index):
        self.requestType = requestType  # type is byte
        self.request = request  # type is byte
        self.value = value  # type is word (= 2 bytes), content request specific
        self.index = index  # type is word
        self.length = None  # type is word, length of the data part
        self.data = []  # type is byte, max length = 1492 bytes

    def getCommandPacket(self, command=None):
        """
        Pack and return the header and the command in a bytes packet.
        """
        if command is None:
            command = ""

        assert type(command) == str

        headerStr = struct.pack(
            ">BBHHH", self.requestType, self.request, self.value, self.index, len(command)
        )
        wrappedCommand = headerStr + command.encode()

        logger.debug(f"Command Packet generated: {wrappedCommand}")

        return wrappedCommand


class PmacEthernetInterface(object):
    """
    This class provides an interface to connect to a remote PMAC over an Ethernet interface.
    It provides methods to connect to the PMAC, to disconnect, and to issue commands.
    """

    def __init__(self, numAxes=None):
        """
        Initialize the PMAC Ethernet interface.
        """
        self.modelName = None  # Will be initialized by the getPmacModel() call

        # Basic connection settings

        self.hostname = ""
        self.port = None

        # Access-to-the-connection semaphore.
        # Use this to lock/unlock I/O access to the connection (whatever type it is) in child classes.

        self.semaphore = threading.Semaphore()

        self.isConnectionOpen = False

        # Use the getter self.getNumberOfAxes() to access this. The value is None if uninitialised.

        if numAxes is not None:
            self._numAxes = int(numAxes)
        else:
            self._numAxes = None

        # Define a number of often used commands

        self.getResponseCommand = EthernetCommand(VR_DOWNLOAD, VR_PMAC_GETRESPONSE, 0, 0)
        self.getBufferCommand = EthernetCommand(VR_UPLOAD, VR_PMAC_GETBUFFER, 0, 0)

    def setConnectionParameters(self, host="localhost", port=None):
        self.hostname = str(host)
        if port is not None:
            self.port = int(str(port))

    # Attempts to open a connection to a remote PMAC.
    # Returns None on success, or an error message string on failure.

    def connect(self):

        # Sanity checks

        if self.isConnectionOpen:
            raise PMACError("Socket is already open")
        if self.hostname in (None, ""):
            raise PMACError("ERROR: hostname not initialized")
        if self.port in (None, 0):
            raise PMACError("ERROR: port number not initialized")

        # Create a new socket instance

        try:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.setblocking(1)
            self.sock.settimeout(3)
        except socket.error as e_socket:
            raise PMACError("ERROR: Failed to create socket.") from e_socket

        # Attempt to establish a connection to the remote host

        try:
            logger.debug(f'Connecting a socket to host "{self.hostname}" using port {self.port}')
            self.sock.connect((self.hostname, self.port))
        except ConnectionRefusedError as e_cr:
            raise PMACError(f"ERROR: Connection refused to {self.hostname}") from e_cr
        except socket.gaierror as e_gai:
            raise PMACError(f"ERROR: socket address info error for {self.hostname}") from e_gai
        except socket.herror as e_h:
            raise PMACError(f"ERROR: socket host address error for {self.hostname}") from e_h
        except socket.timeout as e_timeout:
            raise PMACError(f"ERROR: socket timeout error for {self.hostname}") from e_timeout
        except OSError as ose:
            raise PMACError(f"ERROR: OSError caught ({ose}).") from ose

        self.isConnectionOpen = True

        # Check that we are connected to a pmac by issuing the "VERSION" command -- expecting the 1.947 in return.
        # If we don't get the right response, then disconnect automatically

        if not self.isConnected():
            raise PMACError("Device is not connected, check logging messages for the cause.")

    def disconnect(self):
        """
        Disconnect from the Ethernet connection.

        Raises a PMACError on failure.
        """
        try:
            if self.isConnectionOpen:
                logger.debug(f"Disconnecting from {self.hostname}")
                self.semaphore.acquire()
                self.sock.close()
                self.semaphore.release()
                self.isConnectionOpen = False
        except Exception as e_exc:
            raise PMACError(f"Could not close socket to {self.hostname}") from e_exc

    def isConnected(self):
        """
        Check if the PMAC Ethernet connection is open. This will send a 'version' command to the
        PMAC interface and try to parse the return value. If no connection was established or
        no version could be parsed from the return value, False is returned.
        """
        if not self.isConnectionOpen:
            return False

        try:
            response = self.getResponse(CMD_STRING_VER)
        except PMACError as e_pmac:
            if len(e_pmac.args) >= 2 and e_pmac.args[1] == ERR_CONNECTION:
                logger.error(
                    f"While trying to talk to the device the following exception occurred, exception={e_pmac}"
                )
                logger.error("Most probably the client connection was closed. Disconnecting...")
                self.disconnect()
                return False
            else:
                logger.error(
                    f"While trying to talk to the device the following exception occured, exception={e_pmac}"
                )
                self.disconnect()
                return False
        finally:
            pass

        version = match_float_response(response)

        if not version:
            # if the response was not matching the regex then we're not talking to a PMAC!
            logger.error(
                f'Device did not respond correctly to a "VERSION" command, response={response}. Disconnecting...'
            )
            self.disconnect()
            return False

        return True

    def getBuffer(self, shouldWait=True):
        """"""

        if not self.isConnectionOpen:
            raise PMACError("Device is not connected, reconnect or check logging messages.")

        try:
            if shouldWait:
                self.semaphore.acquire()

            # Attempt to send the complete command to PMAC

            self.sock.sendall(self.getBufferCommand.getCommandPacket())
            logger.debug("Sent out get buffer request to PMAC")

            # wait for, read and return the response from PMAC (will be at most 1400 chars)

            returnStr = self.sock.recv(2048)
            logger.debug(f"Received from PMAC: {returnStr}")

            return returnStr

        except socket.timeout as e_timeout:
            raise PMACError("Socket timeout error", ERR_CMD_TIMEOUT) from e_timeout
        except socket.error as e_socket:
            # Interpret any socket-related error as an I/O error
            raise PMACError("Socket communication error.", ERR_CONNECTION) from e_socket
        except OSError as e_os:
            raise PMACError(f"OS Error: {e_os}") from e_os
        finally:
            if shouldWait:
                self.semaphore.release()

    def getResponse(self, command, shouldWait=True):
        """
        Send a single command to the controller and block until a response from the controller.

        Args:
            command (str): is the command to be sent

            shouldWait (bool, optional, default True): whether to wait on the semaphore.

                This should normally be left default. If we have acquired the semaphore manually,
                then specify shouldWait = False (and don't forget to release the semaphore eventually).

        Returns:
            response (str): is either a string returned by the PMAC (on success), or an error message (on failure)

        Raises:
            PMACError when there was an I/O problem during comm with the PMAC or the response does not have
            recognised terminators.

            Note that PMAC may still return an "ERRxxx" code; this function will still consider that
            a successful transmission.

        """

        if not self.isConnectionOpen:
            raise PMACError("Device is not connected, reconnect or check logging messages.")

        command = str(command)

        try:
            if shouldWait:
                self.semaphore.acquire()

            # Attempt to send the complete command to PMAC

            logger.debug(f"Sending out to PMAC: {command}")
            self.sock.sendall(self.getResponseCommand.getCommandPacket(command))

            # wait for, read and return the response from PMAC (will be at most 1400 chars)

            returnStr = self.sock.recv(2048)
            logger.debug( f"Received from PMAC: {returnStr}")

            return returnStr

        except OSError as e_os:
            raise PMACError(f"OS Error: {e_os}") from e_os
        except socket.timeout as e_timeout:
            raise PMACError("Socket timeout error", ERR_CMD_TIMEOUT) from e_timeout
        except socket.error as e_socket:
            # Interpret any socket-related error as an I/O error
            raise PMACError("Socket communication error.", ERR_CONNECTION) from e_socket
        finally:
            if shouldWait:
                self.semaphore.release()

    def waitFor(self, cmd, values):
        start = datetime.now()
        count = 0

        assert isinstance(values, list), "Expected list argument"

        while datetime.now() - start < DEF_TPMAC_CMD_TIMEOUT:

            # Only read every DEF_TPMAC_READ_FREQ

            if datetime.now() - start > count * DEF_TPMAC_READ_FREQ:
                count += 1
                retStr = self.getResponse(cmd)
                key = match_int_response(retStr)
                if key in values:
                    return key

        raise PMACError(f"Timeout while waiting for {cmd} to be {values}")

    def waitForOutput(self, cmd, values, nr, func):
        start = datetime.now()
        count = 0

        assert isinstance(values, list), "Expected list argument"

        while datetime.now() - start < DEF_TPMAC_CMD_TIMEOUT:

            # Only read every DEF_TPMAC_READ_FREQ

            if datetime.now() - start > count * DEF_TPMAC_READ_FREQ:
                count += 1
                retStr = self.getResponse(cmd)
                Q20, out = extractQ20AndOutput(retStr, nr, func)
                if Q20 not in values:
                    continue
                return out

        raise PMACError(f"Timeout while waiting for {cmd} to be {values}")

    def getQVars(self, base, offsets, func=bytes.decode):
        """
        Get a list of values of Qxx variables, where xx = base + offset (for each offset).

        Throws a PMACError when no response received from Hexapod Hardware Controller.
        """
        qVars = map(lambda x: base + x, offsets)
        cmd = ""
        for qVar in qVars:
            cmd += f"Q{qVar:02} "
        retStr = self.getResponse(cmd)
        logger.debug(f"retStr={retStr} of type {type(retStr)}")

        if retStr == b"\x00":
            raise PMACError(f"No response received for {cmd}, return value is {retStr}")

        # Remove the acknowledgement from the sourceStr

        retStr = retStr.rstrip(b"\r\x06")

        # Split the string in it's parts separated by \r

        qVarsResult = [func(x) for x in retStr.split(b"\r")]

        return qVarsResult

    def getMVars(self, base, offsets, func=bytes.decode):
        """
        Get a list of values of Mxx variables, where xx = base + offset (for each offset).

        The 'M' variables are Symétrie internal variables.
        The following M-Vars are known:

        * M5282 - check if a motion program is running, return 1 when it is running, 0 if no motion program is running

        Throws a PMACError when no response received from Hexapod Hardware Controller.
        """
        mVars = map(lambda x: base + x, offsets)
        cmd = ""
        for mVar in mVars:
            cmd += f"M{mVar:02} "
        retStr = self.getResponse(cmd)
        logger.debug(f"retStr={retStr} of type {type(retStr)}")

        if retStr == b"\x00":
            raise PMACError(f"No response received for {cmd}, return value is {retStr}")

        # Remove the acknowledgement from the sourceStr

        retStr = retStr.rstrip(b"\r\x06")

        # Split the string in it's parts separated by \r

        mVarsResult = [func(x) for x in retStr.split(b"\r")]

        return mVarsResult

    def getIVars(self, base, offsets, func=bytes.decode):
        """
        Get a list of values of Ixxxx variables, where xxxx = base + offset (for each offset).
        """
        iVars = map(lambda x: base + x, offsets)
        cmd = ""
        for iVar in iVars:
            cmd += "i%d " % iVar
        retStr = self.getResponse(cmd)
        logger.debug(f"retStr={retStr} of type {type(retStr)}")

        # Remove the acknowledgement from the sourceStr

        retStr = retStr.rstrip(b"\r\x06")

        # Split the string in it's parts separated by \r

        iVarsResult = [func(x) for x in retStr.split(b"\r")]

        return iVarsResult

    def getPmacModel(self):
        """
        Return a string designating which PMAC model this is.

        Raise a PMACError if the model is unknown.
        """

        if self.modelName:
            return self.modelName

        # Ask for pmac model, returns an integer

        modelCode = self.getCID()

        # Return a model designation based on model code

        modelNamesByCode = {602413: "Turbo PMAC2-VME", 603382: "Geo Brick (3U Turbo PMAC2)"}
        try:
            modelName = modelNamesByCode[modelCode]
        except KeyError as e_key:
            raise PMACError(f"Unsupported PMAC model: modelCode={modelCode}") from e_key

        self.modelName = modelName

        return modelName

    def getCID(self):
        retStr = self.getResponse(CMD_STRING_CID)
        return match_int_response(retStr)

    def getVersion(self):
        retStr = self.getResponse(CMD_STRING_VER)
        return match_float_response(retStr)

    def getCPU(self):
        retStr = self.getResponse(CMD_STRING_CPU)
        return match_string_response(retStr)

    def getType(self):
        retStr = self.getResponse(CMD_STRING_TYPE)
        return match_string_response(retStr)

    def getVID(self):
        retStr = self.getResponse(CMD_STRING_VID)
        return match_string_response(retStr)

    def getDate(self):
        retStr = self.getResponse(CMD_STRING_DATE)
        return match_string_response(retStr)

    def getTime(self):
        retStr = self.getResponse(CMD_STRING_TIME)
        return match_string_response(retStr)

    def getToday(self):
        retStr = self.getResponse(CMD_STRING_TODAY)
        return match_string_response(retStr)

    def sendCommand(self, cmd, **kwargs):
        """
        Sends a command string to the TPMAC.

        Once the acknowledgment is received, it interrogates the TPMAC Q20 variable
        each DEF_TPMAC_READ_FREQ microseconds until Q20 = 0 or -1 or -2.

        Throws a PMACError:
            * when the arguments do not match up
            * when there is a Time-Out
            * when there is a socket communication error

        Returns the Q20 value as an integer.
        """

        # When we expect input parameters, make sure the number of **kwargs matches the cmd['in']
        # and format the command with the given keyword arguments.raise

        if cmd["in"] > 0:
            if len(kwargs) == cmd["in"]:
                fullCommand = cmd["cmd"].format(**kwargs)
            else:
                raise PMACError(
                    f"Expected {cmd['in']} keyword arguments to cmd['name'], got {len(kwargs)}"
                )
        else:
            fullCommand = cmd["cmd"]

        logger.debug(f'Sending the {cmd["name"]} command.')

        retStr = self.getResponse(fullCommand)

        logger.debug(f'Command \'{cmd["name"]}\' returned "{retStr}"')

        # Check the return code (usually Q20)

        if cmd["out"] == 0 and cmd["return"] is not None:

            # The following method can throw a PMACError on Timeout

            rc = self.waitFor(cmd["return"], cmd["check"])
            logger.debug(f"waitFor returned {rc}")

            return rc

        # Get and return the output parameters

        if cmd["out"] > 0 and cmd["return"] is not None:

            # The following method can throw a PMACError on Timeout

            out = self.waitForOutput(cmd["return"], cmd["check"], cmd["out"], cmd["out_type"])
            logger.debug(f"waitForAndOutput returned {out}")

            return out

        # Get the output parameters out of the retStr

        if cmd["out"] > 0 and cmd["return"] is None:
            out = extractOutput(retStr, cmd["out"], cmd["out_type"])

            return out
        #
        return None


if __name__ == "__main__":
    pass



import socket
import threading
import time
from telnetlib import Telnet
from time import sleep
from typing import List

import paramiko

from egse.device import DeviceTransport
from egse.hexapod.symetrie import logger
from egse.system import wait_until


RETURN_CODES = {
    0: "Success.",
    -1: "Undefined error.",
    -10: "Wrong value for parameter at index 0.",
    -11: "Wrong value for parameter at index 1.",
    -12: "Wrong value for parameter at index 2.",
    -13: "Wrong value for parameter at index 3.",
    -14: "Wrong value for parameter at index 4.",
    -15: "Wrong value for parameter at index 5.",
    -16: "Wrong value for parameter at index 6.",
    -17: "Wrong value for parameter at index 7.",
    -18: "Wrong value for parameter at index 8.",
    -19: "Wrong value for parameter at index 9.",
    -20: "Wrong value for parameter at index 10.",
    -21: "Wrong value for parameter at index 11.",
    -22: "Wrong value for parameter at index 12.",
    -23: "Wrong value for parameter at index 13.",
    -24: "Wrong value for parameter at index 14.",
    -25: "Wrong value for parameter at index 15.",
    -26: "Wrong value for parameter at index 16.",
    -27: "Wrong value for parameter at index 17.",
    -28: "Wrong value for parameter at index 18.",
    -29: "Wrong value for parameter at index 19.",
    -30: "Unknown command number.",
    -31: "This configuration command is a 'get' only type.",
    -32: "This configuration command is a 'set' only type.",
    -33: "The axis number do not correspond to an axis defined on the controller.",
    -34: "A stop task is running.",
    -35: "All motors need to be control on.",
    -36: "All motors need to be control off.",
    -37: "Emergency stop is pressed.",
    -38: "A motion task is running.",
    -39: "A home task is running.",
    -40: "Requested move is not feasible.",
    -41: "Power supply of limit switches is off.",
    -42: "Power supply of encoders is off.",
    -43: "A fatal error is present. This type of error needs a controller restart to be removed.",
    -44: "An error is present, error reset is required.",
    -45: "Home is not completed.",
    -46: "Software option not available (can be linked to hardware configuration).",
    -47: "Virtual home: file was created on another controller (different MAC address).",
    -48: "Virtual home: some positions read in file are out of software limits.",
    -49: "Virtual home: file data were stored while hexapod was moving.",
    -50: "Virtual home: no data available.",
    -51: "Command has been rejected because another action is running.",
    -52: "Timeout waiting for home complete status.",
    -53: "Timeout waiting for control on status.",
    -54: "Timeout on motion program start.",
    -55: "Timeout on home task start.",
    -56: "Timeout on virtual home write file task.",
    -57: "Timeout on virtual home delete file task.",
    -58: "Timeout on virtual home read file task.",
    -59: "Timeout on disk access verification task.",
    -60: "Configuration file: save process failed.",
    -61: "Configuration file: loaded file is empty.",
    -62: "Configuration file: loaded data are corrupted.",
    -63: "No access to the memory disk.",
    -64: "File does not exist.",
    -65: "Folder access failed.",
    -66: "Creation of folder tree on the memory disk failed.",
    -67: "Generation or write of the checksum failed.",
    -68: "File read: no data or wrong data size.",
    -69: "File read: no checksum.",
    -70: "File read: incorrect checksum.",
    -71: "File write: failed.",
    -72: "File open: failed.",
    -73: "File delete: failed.",
    -74: "Get MAC address failed.",
    -75: "NaN (Not a Number) or infinite value found.",
    -76: "The coordinate system transformations are not initialized.",
    -77: "A kinematic error is present.",
    -78: "The motor phase process failed (phase search or phase set from position offset).",
    -79: "The motor phase is not found.",
    -80: "Timeout waiting for control off status.",
    -81: "The requested kinematic mode (number) is not defined for the machine.",
    -82: "Timeout waiting for phase found status.",
    -1000: "Internal error: 'RET_Dev_CfS_NaNReturned'.",
    -1001: "Internal error: 'RET_Dev_CfS_FctNotAvailableInKernel'.",
    -1002: "Internal error: 'RET_Dev_CfS_UndefinedCfSType'.",
    -1003: "Internal error: 'RET_Dev_CfS_FIO_UndefinedFioType'.",
    -1004: "Internal error: 'RET_Dev_CfS_FIO_HomeFile_UndefinedAction'.",
    -1005: "Internal error: 'RET_Dev_UndefinedEnumValue'.",
    -1006: "Internal error: 'RET_Dev_LdataCmdStatusIsNegative'.",
    -1007: "Internal error: 'RET_Dev_NumMotorsInCoord_Sup_DEF_aGrQ_SIZE'.",
    -1008: "Internal error: 'RET_Dev_NumMotorsInCoord_WrongNumber'.",
    -1009: "Internal error: 'RET_String_StrCat_DestSizeReached'.",
    -1010: "Internal error: 'RET_String_LengthOverStringSize'.",
    -1011: "Internal error: 'RET_String_AllCharShouldIntBetween_0_255'.",
    -1012: "Internal error: 'RET_String_StrCpy_DestSizeReached'.",
    -1013: "Internal error: 'RET_ErrAction_HomeReset'.",
    -1014: "Internal error: 'RET_Home_StopReceivedWhileRunning'.",
    -1015: "Internal error: 'RET_UndefinedKinAssembly'.",
    -1016: "Internal error: 'RET_WrongPmcConfig'.",
}


class ZondaError(Exception):
    pass


class ZondaSSHInterface:
    def __init__(self, hostname: str):
        self.client = None
        self.gpascii_client = None
        self.connected = False
        self.ssh_output = None
        self.ssh_error = None
        self.verbose = False
        self.hostname = hostname

        self.semaphore = threading.Semaphore()

        self.CommandReturns = RETURN_CODES

    # Etablish the SSH connection with the controller and open gpascii
    def connect(self, ip):
        try:
            # Paramiko.SSHClient can be used to make connections to the remote server and
            # transfer files
            logger.info("Establishing ssh connection with {}...".format(ip))
            self.client = paramiko.SSHClient()
            # Parsing an instance of the AutoAddPolicy to set_missing_host_key_policy() changes
            # it to allow any host.
            self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            self.client.connect(hostname=ip, port=22, username="root", password="deltatau",
                                timeout=5,
                                allow_agent=False, look_for_keys=False)
            logger.warning("Connected to the server: {}".format(ip))
            self.connected = True
        except paramiko.AuthenticationException:
            logger.warning("Authentication failed, please verify your credentials.")
            return False
        except paramiko.SSHException as sshException:
            logger.warning("Could not establish SSH connection: %s" % sshException)
            return False
        except socket.timeout as e:
            logger.warning("Connection timed out.")
            return False
        except Exception as e:
            logger.warning("Exception in connecting to the server.")
            logger.warning("Python says:", e)
            self.client.close()
            return False

        self.gpascii_client = self.client.invoke_shell(term="vt100")
        self.sendcommand("gpascii -2", 0.5)
        self.sendcommand("echo7")

    def is_connected(self):
        return self.connected

    # send a command through gpascii
    def sendcommand(self, command, wait=0):
        self.gpascii_client.send(command + "\r\n")
        if self.verbose == True:
            pass  # print("Command send: {}".format(command))

        if wait > 0:
            sleep(wait)

        # wait for response from gpascii
        step = 0.1
        time = 0

        while True:
            if self.gpascii_client.recv_ready():
                break
            if time > 2.0:
                logger.warning("Command response timed out!")
                return ""
            sleep(step)
            time += step

        # get and clean response
        response = self.gpascii_client.recv(2048)
        response = response.decode()
        response = response.replace(command, "")
        response = response.replace("\r", "")
        # response = response.replace("\n"," ")
        response = response.replace("\x06", "")

        # remove empty elements
        cleaned = ""
        lines = response.split("\n")
        for line in lines:
            if line == "" or line == " ":
                lines.remove(line)
            else:
                cleaned += line + "\n"

        # if self.verbose is True:
        # print("Command rcv: {}".format(cleaned))
        return cleaned

    def cmd_decode(self, name, arguments=list()):
        command = ""

        if "JOG" in name:
            command += "c_ax={} ".format(arguments[0])
            arguments.pop(0)

        # set cfg part
        cfg = 1
        if "?" in name:
            cfg = 0
            name = name.replace("?", "")
        if "CFG_" in name:
            command += "c_cfg={} ".format(cfg)

        # create the parameters part of a command using the given arguments
        for index in range(0, len(arguments)):
            command += "c_par({})={} ".format(index, arguments[index])

        # finish command line
        command += "c_cmd=C_{}".format(name)

        # logger.debug("command sent: ", command)
        return command

    def cmd_check(self):
        # sends "c_cmd" and returns the state of the command that has been previously sent.
        Timer = 0
        last = time.process_time()

        # For a maximum of 5 sec
        while Timer < 5:

            # wait and count
            time.sleep(0.1)
            Timer += time.process_time() - last
            last = time.process_time()

            # ask for command state
            response = self.sendcommand("c_cmd")
            response.replace("\n", "")

            if "=" in response:
                # get value
                elements = response.split("=", 1)
                element = elements[1]

                # exit if command is completed
                if int(element) <= 0:
                    response = element

            code = int(response)

            if code < 0:
                message = self.CommandReturns[code]
                logger.warning("Command Error return: {} : {}".format(code, message))
            elif code == 0:
                message = "Command successful: 0, execution time {} sec".format(Timer)
                # logger.info(message)
            return [code, message]
        logger.warning("Error: Order timed out !")
        return [-1, "Error: Order timed out !"]

    def p_check(self, number=20):
        # ask the value of the command parameters
        command = "c_par(0)"
        if number > 1:
            command += ",{},1".format(number)

        answer = self.sendcommand(command)
        return answer

    def disconnect(self):
        try:
            logger.info(f"Closing ssh connection with {self.hostname}...")
            self.client.close()
            self.gpascii_client.close()
            self.connected = False
            logger.warning("...disconnected from ssh")
        except Exception as e_exc:
            raise ZondaError(f"Could not close socket to {self.hostname}") from e_exc


class ZondaTelnetInterface(DeviceTransport):
    """
    The ZONDA Hexapod device interface based on the telnet protocol.
    """

    TELNET_TIMEOUT = 0.5

    def __init__(self, hostname: str, port: int):
        self.telnet = Telnet()
        self.hostname = hostname
        self.port = port
        self._is_connected = False

    def connect(self):
        self.telnet.open(self.hostname, self.port)
        self.telnet.read_until(b"login: ", timeout=self.TELNET_TIMEOUT)
        self.telnet.write(b"root\r\n")
        self.telnet.read_until(b"Password: ", timeout=self.TELNET_TIMEOUT)
        self.telnet.write(b"deltatau\r\n")
        self.telnet.read_until(b"ppmac# ", timeout=self.TELNET_TIMEOUT)
        self.telnet.write(b"gpascii -2\r\n")
        self.telnet.write(b"echo7\r\n")
        self.telnet.read_until(b'\x06', timeout=self.TELNET_TIMEOUT)
        self.telnet.read_very_eager()
        self._is_connected = True

    def is_connected(self):
        return self._is_connected

    def disconnect(self):
        self.telnet.read_very_eager()
        self.telnet.close()
        self._is_connected = False

    def check_command_status(self):

        if wait_until(lambda: self.trans('c_cmd')[0] == '0', interval=0.01):
            rc = int(self.trans('c_cmd')[0])
            logger.warning(f"Command check: {RETURN_CODES[rc]} [{rc}]")
        else:
            rc = 0

        return rc, RETURN_CODES[rc]

    def get_pars(self, count: int = 20):
        # ask the value of the command parameters
        command = "c_par(0)"
        if count > 1:
            command += f",{count},1"

        return self.trans(command)

    def trans(self, cmd: str) -> List:
        self.write(cmd)
        response = self.read()
        if response and response[0] == cmd:
            return response[1:]
        else:
            return response

    def read(self) -> List:
        response = self.telnet.read_until(b'\x06\r\n', timeout=self.TELNET_TIMEOUT)
        response = response.decode()
        parts = response.split("\r\n")
        if len(parts) == 1:
            return []
        if parts[-2] == '\x06':
            return parts[:-2]
        logger.warning("Expected ACK at the end of the response")
        return parts

    def write(self, cmd: str):
        self.telnet.write(cmd.encode() + b"\r\n")


def decode_command(cmd_name: str, *args):

    args = list(args)
    cmd_name = cmd_name.upper()

    full_command = ""

    if "JOG" in cmd_name:
        arg = args.pop(0)
        full_command += f"c_ax={arg} "

    # set cfg part

    if "?" in cmd_name:
        cmd_name = cmd_name.replace("?", "")
        cfg = 0
    else:
        cfg = 1

    if "CFG_" in cmd_name:
        full_command += f"c_cfg={cfg} "

    # create the parameters part of a full_command using the given arguments

    for index, arg in enumerate(args):
        full_command += f"c_par({index})={arg} "

    # finish full_command line

    full_command += f"c_cmd=C_{cmd_name}"

    return full_command


if __name__ == "__main__":
    import rich

    zonda = ZondaTelnetInterface("192.168.56.10", 23)
    zonda.connect()
    rich.print(zonda.trans("s_hexa,50,1"))
    rich.print(zonda.trans("s_ax_1,6,1"))
    rich.print(zonda.trans("s_pos_ax_1,6,1"))
    rich.print(zonda.read())
    zonda.disconnect()

    print(decode_command("CFG_LIMIT?", 0))
    print(decode_command("CFG_LIMIT", 1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12))



from pathlib import Path

from egse.control import ControlServer
from egse.procman import ProcessManagerController, ProcessManagerCommand
from egse.protocol import CommandProtocol
from egse.settings import Settings
from egse.system import format_datetime
from egse.zmq_ser import bind_address

HERE = Path(__file__).parent

CTRL_SETTINGS = Settings.load("Process Manager Control Server")
COMMAND_SETTINGS = Settings.load(location=HERE, filename="procman.yaml")


class ProcessManagerProtocol(CommandProtocol):

    """
    Command Protocol for Process Management.
    """

    def __init__(self, control_server: ControlServer):
        """ Initialisation of a new Protocol for Process Management.

        The initialisation of this Protocol consists of the following steps:

            - Create a Controller to which the given Control Server should send commands;
            - Load the commands;
            - Build a look-up table for the commands.

        Args:
            - control_server (ControlServer): Control Server via which commands should be sent to the Controller
        """

        super().__init__(control_server)

        # Create a new Controller for Process Management

        self.controller = ProcessManagerController()

        # Load the commands (for commanding of the PM Controller) from the
        # YAML file into a dictionary, stored in the PM Protocol

        self.load_commands(COMMAND_SETTINGS.Commands, ProcessManagerCommand, ProcessManagerController)

        # Build a look-up table for the methods

        self.build_device_method_lookup_table(self.controller)

    def get_bind_address(self):
        """ Returns the address to bind a socket to.

        This bind address is a properly formatted URL, based on the communication protocol and the commanding port.

        Returns: Properly formatted URL to bind a socket to.
        """

        return bind_address(self.control_server.get_communication_protocol(), self.control_server.get_commanding_port())

    def get_status(self) -> dict:
        """ Returns the status information for the Control Server.

        This status information is returned in the form of a dictionary that contains the following information about
        the Control Server for Process Management:

            - timestamp (str): string representation of the current datetime;
            - PID (int): process ID for the Control Server;
            - Up (float): uptime of the Control Server [s];
            - UUID (uuid1): Universally Unique Identifier for the Control Server;
            - RSS (int): 'Resident Set Size', this is the non-swapped physical memory a process has used [byte];
            - USS (int): 'Unique Set Size', this is the memory which is unique to a process [byte];
            - CPU User (float): time spent in user mode [s];
            - CPU System (float): time spent in kernel mode [s];
            - CPU count: number of CPU cores in use by the process;
            - CPU% (float): process CPU utilization as a percentage [%].

        Returns: Dictionary with status information for the Control Server for Process Management.
        """

        return super().get_status()

    def get_housekeeping(self) -> dict:
        """ Returns the housekeeping data for the Control Server.

        This housekeeping data is returns in the form of a dictionary that contains the following information about
        the Control Server for Process Management:

            - timestamp (str): string representation of the current datetime.

        Returns: Dictionary with housekeeping data for the Control Server for Process Management.
        """

        return {"timestamp": format_datetime()}



"""
The Control Server that connects to the Hexapod JORAN Hardware Controller.

Start the control server from the terminal as follows:

    $ joran_cs start-bg

or when you don't have the device available, start the control server in simulator mode. That
will make the control server connect to a device software simulator:

    $ joran_cs start --sim

Please note that software simulators are intended for simple test purposes and will not simulate
all device behavior correctly, e.g. timing, error conditions, etc.

"""
import logging

from egse.process import SubProcess

if __name__ != "__main__":
    import multiprocessing
    multiprocessing.current_process().name = "joran_cs"

import sys

import click
import rich
import zmq

from egse.control import ControlServer
from egse.control import is_control_server_active
from egse.hexapod.symetrie.joran import JoranProxy
from egse.hexapod.symetrie.joran_protocol import JoranProtocol
from egse.settings import Settings
from egse.zmq_ser import connect_address
from prometheus_client import start_http_server

logger = logging.getLogger(__name__)

CTRL_SETTINGS = Settings.load("Hexapod JORAN Control Server")


class JoranControlServer(ControlServer):
    """JoranControlServer - Command and monitor the Hexapod JORAN hardware.

    This class works as a command and monitoring server to control the Symétrie Hexapod JORAN.
    This control server shall be used as the single point access for controlling the hardware
    device. Monitoring access should be done preferably through this control server also,
    but can be done with a direct connection through the PunaController if needed.

    The sever binds to the following ZeroMQ sockets:

    * a REQ-REP socket that can be used as a command server. Any client can connect and
      send a command to the Hexapod.

    * a PUB-SUP socket that serves as a monitoring server. It will send out Hexapod status
      information to all the connected clients every five seconds.

    """

    def __init__(self):
        super().__init__()

        self.device_protocol = JoranProtocol(self)

        self.logger.debug(f"Binding ZeroMQ socket to {self.device_protocol.get_bind_address()}")

        self.device_protocol.bind(self.dev_ctrl_cmd_sock)

        self.poller.register(self.dev_ctrl_cmd_sock, zmq.POLLIN)

    def get_communication_protocol(self):
        return CTRL_SETTINGS.PROTOCOL

    def get_commanding_port(self):
        return CTRL_SETTINGS.COMMANDING_PORT

    def get_service_port(self):
        return CTRL_SETTINGS.SERVICE_PORT

    def get_monitoring_port(self):
        return CTRL_SETTINGS.MONITORING_PORT

    def get_storage_mnemonic(self):
        try:
            return CTRL_SETTINGS.STORAGE_MNEMONIC
        except AttributeError:
            return "JORAN"

    def before_serve(self):
        start_http_server(CTRL_SETTINGS.METRICS_PORT)


@click.group()
def cli():
    pass


@cli.command()
@click.option("--simulator", "--sim", is_flag=True, help="Start the Hexapod Joran Simulator as the backend.")
def start(simulator):
    """Start the Hexapod Joran Control Server."""

    if simulator:

        Settings.set_simulation_mode(True)

    try:

        controller = JoranControlServer()
        controller.serve()

    except KeyboardInterrupt:

        print("Shutdown requested...exiting")

    except SystemExit as exit_code:

        print("System Exit with code {}.".format(exit_code))
        sys.exit(exit_code)

    except Exception:

        logger.exception("Cannot start the Hexapod Joran Control Server")

        # The above line does exactly the same as the traceback, but on the logger
        # import traceback
        # traceback.print_exc(file=sys.stdout)

    return 0


@cli.command()
@click.option("--simulator", "--sim", is_flag=True,
              help="Start the Hexapod Joran Simulator as the backend.")
def start_bg(simulator):
    """Start the JORAN Control Server in the background."""
    sim = "--simulator" if simulator else ""
    proc = SubProcess("joran_cs", ["joran_cs", "start", sim])
    proc.execute()


@cli.command()
def stop():
    """Send a 'quit_server' command to the Hexapod Joran Control Server."""

    try:
        with JoranProxy() as proxy:
            sp = proxy.get_service_proxy()
            sp.quit_server()
    except ConnectionError:
        rich.print("[red]Couldn't connect to 'joran_cs', process probably not running. ")


@cli.command()
def status():
    """Request status information from the Control Server."""

    protocol = CTRL_SETTINGS.PROTOCOL
    hostname = CTRL_SETTINGS.HOSTNAME
    port = CTRL_SETTINGS.COMMANDING_PORT

    endpoint = connect_address(protocol, hostname, port)

    if is_control_server_active(endpoint):
        rich.print("JORAN Hexapod: [green]active")
        with JoranProxy() as joran:
            sim = joran.is_simulator()
            connected = joran.is_connected()
            ip = joran.get_ip_address()
            rich.print(f"type: ALPHA+")
            rich.print(f"mode: {'simulator' if sim else 'device'}{'' if connected else ' not'} connected")
            rich.print(f"hostname: {ip}")
            rich.print(f"commanding port: {port}")
    else:
        rich.print("JORAN Hexapod: [red]not active")


if __name__ == "__main__":

    logging.basicConfig(level=logging.DEBUG, format=Settings.LOG_FORMAT_FULL)

    sys.exit(cli())



"""
This module provides function to load plugins and settings from entry-points.
"""

__all__ = [
    "load_plugins",
    "get_file_infos",
    "entry_points",
    "HierarchicalEntryPoints",
]
import logging
import os
import sys
import textwrap
import traceback
from pathlib import Path

if sys.version_info >= (3, 12):    # Use the standard library version (Python 3.10+)
    from importlib.metadata import entry_points as lib_entry_points
    from importlib.metadata import EntryPoint
else:
    # Fall back to the backport for Python 3.9
    from importlib_metadata import entry_points as lib_entry_points
    from importlib_metadata import EntryPoint

import click
import rich

_LOGGER = logging.getLogger(__name__)


def entry_points(group: str) -> set[EntryPoint]:
    """
    Returns a set with all entry-points for the given group name.

    When the name is not known as an entry-point group, an empty set will be returned.
    """

    try:
        x = lib_entry_points().select(group=group)
        return {ep for ep in x}  # use of set here to remove duplicates
    except KeyError:
        return set()


def load_plugins(entry_point: str) -> dict:
    """
    Returns a dictionary with plugins loaded. The keys are the names of the entry-points,
    the values are the loaded modules or objects.

    Note:
        When an entry point cannot be loaded, an error is logged and the value for that
        entry point in the returned dictionary will be None.
    """
    eps = {}
    for ep in entry_points(entry_point):
        try:
            eps[ep.name] = ep.load()
        except Exception as exc:
            eps[ep.name] = None
            _LOGGER.error(f"Couldn't load entry point: {exc}")

    return eps


def get_file_infos(entry_point: str) -> dict[str, tuple[Path, str]]:
    """
    Returns a dictionary with location and filename of all the entries found for
    the given entry-point name.

    The entry-points are interpreted as follows: `<name> = "<module>:<filename>"` where

    - `<name>` is the name of the entry-point given in the pyproject.toml file
    - `<module>` is a valid module name that can be imported and from which the location can be determined.
    - `<filename>` is the name of the target file, e.g. a YAML file

    As an example, for the `cgse-common` settings, the following entry in the `pyproject.toml`:

        [project.entry-points."cgse.settings"]
        cgse-common = "cgse_common:settings.yaml"

    Note that the module name for this entry point has an underscore instead of a dash.

    Return:
        A dictionary with the entry point name as the key and a tuple (location, filename) as the value.
    """
    from egse.system import get_module_location

    eps = dict()

    for ep in entry_points(entry_point):
        try:
            path = get_module_location(ep.module)

            if path is None:
                _LOGGER.error(
                    f"The entry-point '{ep.name}' is ill defined. The module part doesn't exist or is a "
                    f"namespace. No settings are loaded for this entry-point."
                )
            else:
                eps[ep.name] = (path, ep.attr)

        except Exception as exc:
            _LOGGER.error(f"The entry point '{ep.name}' is ill defined: {exc}")

    return eps


class HierarchicalEntryPoints:
    def __init__(self, base_group: str):
        self.base_group = base_group
        self._discover_groups()

    def _discover_groups(self):
        """Discover all groups that match the base group pattern."""
        all_eps = lib_entry_points()
        # rich.print(f"{type(all_eps) = }")

        self.groups = {}
        self.flat_entries = []

        for ep in all_eps:
            # rich.print(f"{type(ep) = }, {dir(ep) = }")
            if ep.group == self.base_group or ep.group.startswith(f"{self.base_group}."):
                self.groups[ep.group] = ep
                self.flat_entries.append(ep)

    def get_all_entry_points(self):
        """Get all entry points as a flat list."""
        return self.flat_entries

    def get_by_subgroup(self, subgroup=None):
        """Get entry points from a specific subgroup."""
        if subgroup is None:
            return self.groups.get(self.base_group, [])

        full_group = f"{self.base_group}.{subgroup}"
        return self.groups.get(full_group, [])

    def get_all_groups(self):
        """Get all discovered group names."""
        return list(self.groups.keys())

    def get_by_type(self, entry_type):
        """Get entry points by type (assuming type is the subgroup name)."""
        return self.get_by_subgroup(entry_type)


# The following code was adapted from the inspiring package click-plugins
# at https://github.com/click-contrib/click-plugins/


def handle_click_plugins(plugins):
    def decorator(group):
        if not isinstance(group, click.Group):
            raise TypeError("Plugins can only be attached to an instance of click.Group()")

        for entry_point in plugins or ():
            try:
                group.add_command(entry_point.load())
            except Exception:
                # Catch this so a busted plugin doesn't take down the CLI.
                # Handled by registering a dummy command that does nothing
                # other than explain the error.
                group.add_command(BrokenCommand(entry_point.name))

        return group

    return decorator


# This class is filtered and will not be included in the API docs, see `mkdocs.yml`.


class BrokenCommand(click.Command):
    """
    Rather than completely crash the CLI when a broken plugin is loaded, this
    class provides a modified help message informing the user that the plugin is
    broken, and they should contact the owner.  If the user executes the plugin
    or specifies `--help` a traceback is reported showing the exception the
    plugin loader encountered.
    """

    def __init__(self, name):
        """
        Define the special help messages after instantiating a `click.Command()`.
        """

        click.Command.__init__(self, name)

        util_name = os.path.basename(sys.argv and sys.argv[0] or __file__)
        icon = "\u2020"

        self.help = textwrap.dedent(
            f"""\
            Warning: entry point could not be loaded. Contact its author for help.

            {traceback.format_exc()}
            """
        )

        self.short_help = f"{icon} Warning: could not load plugin. See `{util_name} {self.name} --help`."

    def invoke(self, ctx):
        """
        Print the traceback instead of doing nothing.
        """

        rich.print()
        rich.print(self.help)
        ctx.exit(1)

    def parse_args(self, ctx, args):
        return args



import asyncio
import sys

import rich


async def run_all_status(full: bool = False, suppress_errors: bool = True):
    tasks = [
        asyncio.create_task(status_rm_cs(suppress_errors)),
        asyncio.create_task(status_log_cs(suppress_errors)),
        asyncio.create_task(status_sm_cs(full, suppress_errors)),
        asyncio.create_task(status_cm_cs(suppress_errors)),
        asyncio.create_task(status_pm_cs(suppress_errors))
    ]

    await asyncio.gather(*tasks)


async def status_rm_cs(suppress_errors):

    proc = await asyncio.create_subprocess_exec(
        sys.executable, '-m', 'egse.registry.server', 'status',
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )

    stdout, stderr = await proc.communicate()

    rich.print(stdout.decode().rstrip())
    if stderr and not suppress_errors:
        rich.print(f"[red]{stderr.decode()}[/]")


async def status_log_cs(suppress_errors):

    proc = await asyncio.create_subprocess_exec(
        sys.executable, '-m', 'egse.logger.log_cs', 'status',
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )

    stdout, stderr = await proc.communicate()

    rich.print(stdout.decode().rstrip())
    if stderr and not suppress_errors:
        rich.print(f"[red]{stderr.decode()}[/]")


async def status_sm_cs(full: bool = False, suppress_errors: bool = True):

    cmd = [sys.executable, '-m', 'egse.storage.storage_cs', 'status']
    if full:
        cmd.append("--full" if full else "")

    proc = await asyncio.create_subprocess_exec(
        *cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )

    stdout, stderr = await proc.communicate()

    rich.print(stdout.decode().rstrip())
    if stderr and not suppress_errors:
        rich.print(f"[red]{stderr.decode()}[/]")


async def status_cm_cs(suppress_errors):

    proc = await asyncio.create_subprocess_exec(
        sys.executable, '-m', 'egse.confman.confman_cs', 'status',
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )

    stdout, stderr = await proc.communicate()

    rich.print(stdout.decode().rstrip())
    if stderr and not suppress_errors:
        rich.print(f"[red]{stderr.decode()}[/]")


async def status_pm_cs(suppress_errors):

    proc = await asyncio.create_subprocess_exec(
        sys.executable, '-m', 'egse.procman.procman_cs', 'status',
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )

    stdout, stderr = await proc.communicate()

    rich.print(stdout.decode().rstrip())
    if stderr and not suppress_errors:
        rich.print(f"[red]{stderr.decode()}[/]")



"""
This module provides a dummy implementation for classes of the commanding chain.

Start the control server with:
```shell
py -m egse.dummy start-cs
```
and stop the server with:
```shell
py -m egse.dummy stop-cs
```
Commands that can be used with the proxy:

  * info – returns an info message from the dummy device, e.g. "Dummy Device <__version__>"
  * get_value – returns a random float between 0.0 and 1.0
  * division – returns the result of the division between arguments 'a' and 'b'.
    This can be used also to induce a ZeroDivisionError that should return a Failure
    object.

The device simulator can be started with:
```
py -m egse.dummy start-dev
```
"""

from __future__ import annotations

import contextlib
import logging
import multiprocessing
import random
import select
import socket
import sys
import textwrap
import threading
import time
from functools import partial

import typer
import zmq

from egse.command import ClientServerCommand
from egse.control import ControlServer
from egse.control import is_control_server_active
from egse.decorators import dynamic_interface
from egse.device import DeviceConnectionError
from egse.device import DeviceConnectionInterface
from egse.device import DeviceTimeoutError
from egse.device import DeviceTransport
from egse.listener import Event
from egse.listener import EventInterface
from egse.protocol import CommandProtocol
from egse.proxy import Proxy
from egse.response import Failure
from egse.settings import Settings
from egse.system import SignalCatcher
from egse.system import attrdict
from egse.system import format_datetime
from egse.zmq_ser import bind_address
from egse.zmq_ser import connect_address

logging.basicConfig(level=logging.DEBUG, format=Settings.LOG_FORMAT_FULL)

_LOGGER = logging.getLogger("egse.dummy")

__version__ = "0.2.0"

DEV_HOST = "localhost"
"""The hostname or IP address of the Dummy Device."""
DEV_PORT = 4446
"""The port number for the Dummy Device."""
DEV_NAME = "Dummy Device"
"""The name used for theDummy Device, this is used in Exceptions and in the info command."""

READ_TIMEOUT = 10.0
"""The maximum time in seconds to wait for a socket receive command."""
WRITE_TIMEOUT = 1.0
"""The maximum time in seconds to wait for a socket send command."""
CONNECT_TIMEOUT = 3.0
"""The maximum time in seconds to wait for establishing a socket connect."""

# Especially DummyCommand and DummyController need to be defined in a known module
# because those objects are pickled and when de-pickled at the clients side the class
# definition must be known.

# We use AttributeDict here to define the settings, because that is how the Settings.load() returns
# settings loaded from a YAML file.

ctrl_settings = attrdict(
    {
        "HOSTNAME": "localhost",
        "COMMANDING_PORT": 4443,
        "SERVICE_PORT": 4444,
        "MONITORING_PORT": 4445,
        "PROTOCOL": "tcp",
        "TIMEOUT": 10_000,  # milliseconds
        "HK_DELAY": 1.0,
    }
)

commands = attrdict(
    {
        "info": {"description": "Info on the Dummy Device.", "response": "handle_device_method"},
        "get_value": {
            "description": "Read a value from the device.",
        },
        "division": {
            "description": "Return a / b",
            "cmd": "{a} {b}"
        },
        "handle_event": {
            "description": "Notification of an event",
            "device_method": "handle_event",
            "cmd": "{event}",
            "response": "handle_device_method",
        },
    }
)


def is_dummy_cs_active() -> bool:
    """Returns True if the dummy device control server is active."""
    return is_control_server_active(
        endpoint=connect_address(ctrl_settings.PROTOCOL, ctrl_settings.HOSTNAME, ctrl_settings.COMMANDING_PORT)
    )


class DummyCommand(ClientServerCommand):
    """The Command class for the dummy device."""

    pass


class DummyInterface:
    """The interface for the dummy device."""

    @dynamic_interface
    def info(self) -> str:
        """Return an info string from the device."""
        raise NotImplementedError("The info() method has not been loaded from the service.")

    @dynamic_interface
    def get_value(self, *args, **kwargs) -> float:
        """
        Return a float value from the device.
        This dummy implementation will return a random number between 0.0 and 1.0.
        """
        raise NotImplementedError("The get_value() method has not been loaded from the service.")

    @dynamic_interface
    def division(self, a: int | float, b: int | float) -> float:
        """
        Return the division of the number 'a' divided by the number 'b'.
        This method can also be used during testing to cause a ZeroDivisionError
        that will return a Failure object.
        """
        raise NotImplementedError("The division() method has not been loaded from the service.")


class DummyProxy(Proxy, DummyInterface, EventInterface):
    """
    The Proxy class for the dummy device.

    Args:
        protocol: the transport protocol [default is taken from settings file]
        hostname: location of the control server (IP address) [default is taken from settings file]
        port: TCP port on which the control server is listening for commands [default is taken from settings file]
        timeout: a socket timeout in milliseconds
    """

    def __init__(
            self,
            protocol: str = ctrl_settings.PROTOCOL,
            hostname: str = ctrl_settings.HOSTNAME,
            port: int = ctrl_settings.COMMANDING_PORT,
            timeout: int = ctrl_settings.TIMEOUT,
    ):
        super().__init__(connect_address(protocol, hostname, port), timeout=timeout)


class DummyController(DummyInterface, EventInterface):
    """
    The controller class for the dummy device.

    This class is used to directly communicate with the device.
    """

    def __init__(self, control_server):
        self._cs = control_server
        self._dev = DummyDeviceEthernetInterface(DEV_HOST, DEV_PORT)
        self._dev.connect()

    def info(self) -> str:
        return self._dev.trans("info").decode().strip()

    def get_value(self) -> float:
        return float(self._dev.trans("get_value").decode().strip())

    def division(self, a, b) -> float:
        return a / b

    def handle_event(self, event: Event) -> str:
        _exec_in_thread = False

        def _handle_event(_event):
            _LOGGER.info(f"An event is received, {_event=}")
            _LOGGER.info(f"CM CS active? {is_control_server_active()}")
            time.sleep(5.0)
            _LOGGER.info(f"CM CS active? {is_control_server_active()}")
            _LOGGER.info(f"An event is processed, {_event=}")

        if _exec_in_thread:
            # We execute this function in a daemon thread so the acknowledgment is
            # sent back immediately (the ACK means 'command received and will be
            # executed').

            retry_thread = threading.Thread(target=_handle_event, args=(event,))
            retry_thread.daemon = True
            retry_thread.start()
        else:
            # An alternative to the daemon thread is to create a scheduled task that will be executed
            # after the event is acknowledged.

            self._cs.schedule_task(partial(_handle_event, event))

        return "ACK"


class DummyProtocol(CommandProtocol):
    """
    The protocol class for the dummy device.

    This class defines the communication between the client (usually a Proxy) and
    the server (the control server) for this device.

    Args:
        control_server: the control server for the dummy device.
    """

    def __init__(self, control_server: ControlServer):
        super().__init__(control_server)

        self.device_controller = DummyController(control_server)

        self.load_commands(commands, DummyCommand, DummyController)

        self.build_device_method_lookup_table(self.device_controller)

        self._count = 0

    def get_bind_address(self):
        return bind_address(self.control_server.get_communication_protocol(), self.control_server.get_commanding_port())

    def get_status(self):
        return super().get_status()

    def get_housekeeping(self) -> dict:
        # _LOGGER.debug(f"Executing get_housekeeping function for {self.__class__.__name__}.")

        self._count += 1

        # use the sleep to test the responsiveness of the control server when even this get_housekeeping function takes
        # a lot of time, i.e. up to several minutes in the case of data acquisition devices
        # import time
        # time.sleep(2.0)

        return {
            "timestamp": format_datetime(),
            "COUNT": self._count,
            "PI": 3.14159,  # just to have a constant parameter
            "Random": random.randint(0, 100),  # just to have a variable parameter
        }

    def quit(self):
        _LOGGER.info("Executing 'quit()' on DummyProtocol.")


class DummyControlServer(ControlServer):
    """
    DummyControlServer - Command and monitor dummy device controllers.

    The sever binds to the following ZeroMQ sockets:

    * a REQ-REP socket that can be used as a command server. Any client can connect and
      send a command to the dummy device controller.

    * a PUB-SUP socket that serves as a monitoring server. It will send out status
      information to all the connected clients every DELAY seconds.

    """

    def __init__(self):
        multiprocessing.current_process().name = "dummy_cs"

        super().__init__()

        self.device_protocol = DummyProtocol(self)

        _LOGGER.info(
            f"Binding ZeroMQ socket to {self.device_protocol.get_bind_address()} for {self.__class__.__name__}"
        )

        self.device_protocol.bind(self.dev_ctrl_cmd_sock)

        self.poller.register(self.dev_ctrl_cmd_sock, zmq.POLLIN)

        self.set_hk_delay(ctrl_settings.HK_DELAY)

        try:
            from egse.confman import ConfigurationManagerProxy
            from egse.listener import EVENT_ID

            # The following import is needed because without this import, DummyProxy would
            # be <class '__main__.DummyProxy'> instead of `egse.dummy.DummyProxy` and the
            # ConfigurationManager control server will not be able to de-pickle
            # the register message.

            from egse.dummy import DummyProxy  # noqa

            self.register_as_listener(
                proxy=ConfigurationManagerProxy,
                listener={"name": "Dummy CS", "proxy": DummyProxy, "event_id": EVENT_ID.SETUP},
            )

        except ModuleNotFoundError as exc:
            _LOGGER.info(
                textwrap.dedent(
                    f"""\
                    Caught a ModuleNotFoundException: {exc}

                    This probably means you have not installed the `cgse-core` package and you don't have a
                    configuration manager running. The DummyControlServer will not be registered as a listener and
                    therefore will not receive notifications. Other then that, it should be fully functional.
                    """
                )
            )

    def get_communication_protocol(self):
        return "tcp"

    def get_commanding_port(self):
        return ctrl_settings.COMMANDING_PORT

    def get_service_port(self):
        return ctrl_settings.SERVICE_PORT

    def get_monitoring_port(self):
        return ctrl_settings.MONITORING_PORT

    def get_storage_mnemonic(self):
        return "DUMMY-HK"

    def after_serve(self):
        _LOGGER.debug("After Serve: unregistering Dummy CS")

        with contextlib.suppress(ModuleNotFoundError):
            from egse.confman import ConfigurationManagerProxy
            self.unregister_as_listener(proxy=ConfigurationManagerProxy, listener={"name": "Dummy CS"})


class DummyDeviceEthernetInterface(DeviceConnectionInterface, DeviceTransport):
    """
    Defines the low-level interface to the Dummy Device.

    Args:
        hostname (str): the IP address or fully qualified hostname of the Dummy Device
            controller.

        port (int): the IP port number to connect to.
    """

    def __init__(self, hostname: str = None, port: int = None):
        super().__init__()

        # Basic connection settings, loaded from the configuration YAML file

        self.hostname = hostname
        self.port = port
        self.sock = None

        self.is_connection_open = False

    def connect(self):
        """
        Connects the TCP socket to the device controller.

        Raises:
            ValueError: when hostname or port number are not initialized properly.
            DeviceConnectionError: on any socket error except timeouts.
            DeviceTimeoutError: on a socket timeout.
        """
        # Sanity checks

        if self.is_connection_open:
            _LOGGER.warning("Trying to connect to an already connected socket.")
            return

        if self.hostname in (None, ""):
            raise ValueError("hostname is not initialized.")

        if self.port in (None, 0):
            raise ValueError("port number is not initialized.")

        # Create a new socket instance

        try:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        except socket.error as exc:
            self.sock.close()
            raise DeviceConnectionError("Dummy Device", "Failed to create socket.") from exc

        # Attempt to establish a connection to the remote host

        # Socket shall be closed on exception?
        # Closing a socket doesn't harm and almost never throws an exception,
        # except for fatal exceptions like MemoryError, ...

        # We set a timeout of 3 sec before connecting and reset to None
        # (=blocking) after the connect() method. The reason for this that when no
        # device is available, e.g. during testing, the timeout will take about
        # two minutes which is way too long. It needs to be evaluated if this
        # approach is acceptable and not causing problems during production.

        try:
            _LOGGER.debug(f'Connecting a socket to host "{self.hostname}" using port {self.port}')
            self.sock.settimeout(CONNECT_TIMEOUT)
            self.sock.connect((self.hostname, self.port))
            self.sock.settimeout(None)
        except ConnectionRefusedError as exc:
            self.sock.close()
            raise DeviceConnectionError(DEV_NAME, f"Connection refused to {self.hostname}:{self.port}.") from exc
        except TimeoutError as exc:
            self.sock.close()
            raise DeviceTimeoutError(DEV_NAME, f"Connection to {self.hostname}:{self.port} timed out.") from exc
        except socket.gaierror as exc:
            self.sock.close()
            raise DeviceConnectionError(DEV_NAME, f"socket address info error for {self.hostname}") from exc
        except socket.herror as exc:
            self.sock.close()
            raise DeviceConnectionError(DEV_NAME, f"socket host address error for {self.hostname}") from exc
        except OSError as exc:
            self.sock.close()
            raise DeviceConnectionError(DEV_NAME, f"OSError caught ({exc}).") from exc

        self.is_connection_open = True

        # The first thing to receive should be the device info string.
        # This might not be the case for your device.

        response = self.read()
        _LOGGER.debug(f"After connection, we got '{response.decode().rstrip()}' as a response.")

    def disconnect(self):
        """
        Disconnect the Ethernet connection from the device controller.

        Raises:
             DeviceConnectionError: on failure.
        """
        try:
            _LOGGER.debug(f"Disconnecting from {self.hostname}")
            self.sock.close()
            self.is_connection_open = False
        except Exception as exc:
            raise DeviceConnectionError(DEV_NAME, f"Could not close socket to {self.hostname}") from exc

    def reconnect(self):
        """Disconnect from the device, then connect again."""
        if self.is_connection_open:
            self.disconnect()
        self.connect()

    def is_connected(self) -> bool:
        """
        Check if the device is connected.

        Returns:
             True is the device is connected, False otherwise.
        """

        return bool(self.is_connection_open)

    def read(self) -> bytes:
        """
        Read a response from the device.

        Returns:
            A bytes object containing the response from the device. No processing is done
            on the response.

        Raises:
            DeviceTimeoutError: when the read operation timed out.
        """
        n_total = 0
        buf_size = 1024 * 10
        response = bytes()

        # Set a timeout of READ_TIMEOUT to the socket.recv

        saved_timeout = self.sock.gettimeout()
        self.sock.settimeout(READ_TIMEOUT)

        try:
            for _ in range(100):
                # time.sleep(0.1)  # Give the device time to fill the buffer
                data = self.sock.recv(buf_size)
                n = len(data)
                n_total += n
                response += data
                if n < buf_size:
                    break
        except socket.timeout as exc:
            _LOGGER.warning(f"Socket timeout error: {exc}")
            raise DeviceTimeoutError(DEV_NAME, "Socket timeout error") from exc
        finally:
            self.sock.settimeout(saved_timeout)

        # _LOGGER.debug(f"Total number of bytes received is {n_total}, idx={idx}")
        # _LOGGER.debug(f"> {response[:80]=}")

        return response

    def write(self, command: str) -> None:
        """
        Send a command to the device.

        No processing is done on the command string, except for the encoding into a bytes object.

        Args:
            command: the command string including terminators.

        Raises:
            DeviceTimeoutError: when the sendall() timed out, and a DeviceConnectionError if
                there was a socket related error.
        """

        # _LOGGER.debug(f"{command.encode() = }")

        try:
            self.sock.sendall(command.encode())
        except socket.timeout as exc:
            raise DeviceTimeoutError(DEV_NAME, "Socket timeout error") from exc
        except socket.error as exc:
            # Interpret any socket-related error as an I/O error
            raise DeviceConnectionError(DEV_NAME, "Socket communication error.") from exc

    def trans(self, command: str) -> bytes:
        """
        Send a command to the device and wait for the response.

        No processing is done on the command string, except for the encoding into a bytes object.

        Args:
            command: the command string including terminators.

        Returns:
            A bytes object containing the response from the device. No processing is done
            on the response.

        Raises:
            DeviceTimeoutError: when the sendall() timed out, and a DeviceConnectionError if
                there was a socket related error.
        """
        # MODULE_LOGGER.debug(f"{command.encode() = }")

        try:
            # Attempt to send the complete command

            self.sock.sendall(command.encode())

            # wait for, read and return the response (will be at most TBD chars)

            return self.read()

        except socket.timeout as exc:
            raise DeviceTimeoutError(DEV_NAME, "Socket timeout error") from exc
        except socket.error as exc:
            # Interpret any socket-related error as an I/O error
            raise DeviceConnectionError(DEV_NAME, "Socket communication error.") from exc


app = typer.Typer()


@app.command()
def start_cs():
    """Start the dummy control server on localhost."""

    # The following import is needed because without this import, the control server and Proxy will not be able to
    # instantiate classes that are passed in ZeroMQ messages and de-pickled.
    from egse.dummy import DummyControlServer  # noqa

    try:
        control_server = DummyControlServer()
        control_server.serve()
    except KeyboardInterrupt:
        print("Shutdown requested...exiting")
    except SystemExit as exit_code:
        print(f"System Exit with code {exit_code}.")
        sys.exit(-1)
    except Exception:  # noqa
        import traceback

        traceback.print_exc(file=sys.stdout)


@app.command()
def stop_cs():
    """Send a quit service command to the dummy control server."""
    with DummyProxy() as dummy:
        _LOGGER.info("Sending quit_server() to Dummy CS.")
        sp = dummy.get_service_proxy()
        sp.quit_server()


# ----- Dummy Device functions -----------------------------------------------------------------------------------------

error_msg = ""


@app.command()
def start_dev():
    """Start the dummy device simulator."""
    global error_msg

    multiprocessing.current_process().name = "dummy_dev"

    _LOGGER.info("Starting the Dummy Device simulator")

    quit_request = False
    timeout = 1.0

    killer = SignalCatcher()

    while not quit_request:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind((DEV_HOST, DEV_PORT))
            s.listen()
            conn, addr = s.accept()
            with conn:
                _LOGGER.info(f"Accepted connection from {addr}")
                conn.sendall(f"Dummy Device {__version__}".encode())
                try:
                    while True:
                        error_msg = ""
                        read_sockets, _, _ = select.select([conn], [], [], timeout)
                        if conn in read_sockets:
                            data = conn.recv(4096)
                            if not data:
                                _LOGGER.info("Connection closed by peer, waiting for connection..")
                                break  # connection closed by peer
                            if (response := process_command(data.decode().rstrip())) is not None:
                                response = f"{response}\r\n".encode()
                                conn.sendall(response)
                            _LOGGER.debug(f"{data = } -> {response = }")
                        if killer.term_signal_received:
                            _LOGGER.info("TERM signal received, terminating...")
                            quit_request = True
                            break
                        # _LOGGER.debug(f"Timeout received after {timeout}s..")
                except KeyboardInterrupt:
                    _LOGGER.info("Keyboard interrupt, closing.")
                except ConnectionResetError as exc:
                    _LOGGER.info(f"ConnectionResetError: {exc}")
                except Exception as exc:
                    _LOGGER.info(f"{exc.__class__.__name__} caught: {exc.args}")

    _LOGGER.info("Dummy Device terminated.")


COMMAND_ACTIONS_RESPONSES = {"info": (None, f"Dummy Device {__version__}"), "get_value": (None, random.random)}


def process_command(command_string: str) -> str | None:
    _LOGGER.debug(f"{command_string = }")

    try:
        action, response = COMMAND_ACTIONS_RESPONSES[command_string]
        action and action()
        if error_msg:
            return error_msg
        else:
            return response if isinstance(response, str) else response()
    except KeyError:
        from egse.system import get_caller_breadcrumbs

        _LOGGER.info(get_caller_breadcrumbs())
        raise NotImplementedError(f"{command_string} not yet implemented..")


if __name__ == "__main__":
    app()



from __future__ import annotations

import json
import pickle
import zlib
from enum import IntEnum

import zmq


def connect_address(transport: str, address: str, port: int) -> str:
    """Returns a properly formatted URL to connect to."""
    return f"{transport}://{address}:{port}"


def bind_address(transport: str, port: int) -> str:
    """Returns a properly formatted url to bind a socket to."""
    return f"{transport}://*:{port}"


def set_address_port(url: str, port: int) -> str:
    """Returns a url where the 'port' part is replaced with the given port."""
    transport, address, old_port = split_address(url)

    return f"{transport}://{address}:{port}"


def split_address(url: str) -> tuple[str, str, int]:
    transport, address, port = url.split(":")
    if address.startswith("//"):
        address = address[2:]
    return transport, address, int(port)


def send_zipped_pickle(socket, obj, flags=0, protocol=-1):
    """pickle an object, and zip the pickle before sending it"""
    p = pickle.dumps(obj, protocol)
    z = zlib.compress(p)
    return socket.send(z, flags=flags)


def recv_zipped_pickle(socket, flags=0, protocol=-1):
    """inverse of send_zipped_pickle"""
    z = socket.recv(flags)
    p = zlib.decompress(z)
    return pickle.loads(p)


def get_port_number(socket: zmq.Socket) -> int | None:
    """
    Returns the port number associated with this socket. Returns None for sockets
    that do not bind to a TCP or IPC transport.
    """
    endpoint = socket.getsockopt(zmq.LAST_ENDPOINT)
    if endpoint:
        port = endpoint.decode("utf-8").split(':')[-1]
        return int(port)
    else:
        return None


def zmq_string_request(request: str) -> list:
    return [
        b"MESSAGE_TYPE:STRING",
        request.encode('utf-8'),
    ]


def zmq_string_response(message: str) -> list:
    return [
        b"MESSAGE_TYPE:STRING",
        message.encode('utf-8'),
    ]


def zmq_json_request(request: dict) -> list:
    return [
        b"MESSAGE_TYPE:JSON",
        json.dumps(request).encode(),
    ]


def zmq_json_response(message: dict) -> list:
    return [
        b"MESSAGE_TYPE:JSON",
        json.dumps(message).encode(),
    ]


def zmq_error_response(message: dict) -> list:
    return [
        b"MESSAGE_TYPE:ERROR",
        pickle.dumps(message),
    ]


class MessageIdentifier(IntEnum):
    """
    The first item in a multipart message that can be used to subscribe, filter and identify
    messages.
    """

    # ALL shall not be used in the multipart message itself, but exists as an indicator for
    # subscribing to all messages. The ALL shall be converted into b'' when subscribing.

    ALL = 0x00

    # Synchronisation to DPU Processor at time of reception

    SYNC_TIMECODE = 0x80
    SYNC_HK_PACKET = 0x81
    SYNC_DATA_PACKET = 0x82
    SYNC_ERROR_FLAGS = 0x85
    SYNC_HK_DATA = 0x86

    N_FEE_REGISTER_MAP = 0x83
    NUM_CYCLES = 0x84

    # Sending out all kinds of information

    HDF5_FILENAMES = 0x90



import os
import sys

import typer

app = typer.Typer(name="scratch")

os.environ['SKIPPY'] = 'skip'

import check_import  # noqa

print(f"{sys.argv = }")


@app.command()
def start():

    print(f"{sys.argv = }")
    print(f"{os.getenv('SKIPPY') = }")


@app.command()
def end():

    print("Endgame...")


if __name__ == '__main__':
    app()



import nox

python_versions = ['3.9', '3.10', '3.11', '3.12', '3.13']


# Define monorepo-wide sessions

@nox.session
def ruff(session):
    """Ruff the entire codebase."""
    session.install("ruff")
    session.run("ruff", "check", "--no-fix", "libs/cgse-common/src")
    session.run("ruff", "check", "--no-fix", "libs/cgse-core/src")



import contextlib
import logging
import re
import subprocess
import sys
import textwrap
from pathlib import Path
from typing import Annotated

import rich
import typer

from cgse_common import AppState
from egse.plugin import entry_points
from egse.setup import Setup
from egse.system import format_datetime

app = typer.Typer()


@app.command()
def top(ctx: typer.Context):
    """
    A top-like interface for core services and device control servers.

    Not yet implemented.
    """
    state: AppState = ctx.obj

    print(f"The `verbose` flag is {state.verbose}")
    print("This fancy top is not yet implemented.")


@app.command()
def init(project: str = ""):
    """Initialize your project."""
    from rich.prompt import Prompt, Confirm

    project = project.upper()
    site_id = None

    rich.print("[light_steel_blue]Please note default values are given between \[brackets].[/]")

    answer = Prompt.ask(f"What is the name of the project [{project}] ?")
    if answer:
        project = answer.upper()
    while not site_id:
        answer = Prompt.ask("What is the site identifier ?")
        if answer:
            site_id = answer.upper()
        else:
            answer = Confirm.ask("Abort?")
            if answer:
                return

    data_storage_location = f"~/data/{project}/{site_id}/"
    answer = Prompt.ask(f"Where can the project data be stored [{data_storage_location}] ?")
    if answer:
        data_storage_location = answer

    conf_data_location = f"~/data/{project}/{site_id}/conf/"
    answer = Prompt.ask(f"Where will the configuration data be located [{conf_data_location}] ?")
    if answer:
        conf_data_location = answer

    log_file_location = f"~/data/{project}/{site_id}/log/"
    answer = Prompt.ask(f"Where will the logging messages be stored [{log_file_location}] ?")
    if answer:
        log_file_location = answer

    local_settings_path = f"~/data/{project}/{site_id}/local_settings.yaml"
    answer = Prompt.ask(f"Where shall I create a local settings YAML file [{local_settings_path}] ?")
    if answer:
        local_settings_path = answer

    Path(data_storage_location).expanduser().mkdir(exist_ok=True, parents=True)
    (Path(data_storage_location).expanduser() / "daily").mkdir(exist_ok=True, parents=True)
    (Path(data_storage_location).expanduser() / "obs").mkdir(exist_ok=True, parents=True)
    Path(conf_data_location).expanduser().mkdir(exist_ok=True, parents=True)
    Path(log_file_location).expanduser().mkdir(exist_ok=True, parents=True)

    if not Path(local_settings_path).expanduser().exists():
        with open(Path(local_settings_path).expanduser(), 'w') as fd:
            fd.write(
                textwrap.dedent(
                    f"""\
                    SITE:
                        ID:  {site_id}
                    """
                )
            )

    answer = Confirm.ask("Shall I add the environment to your ~/bash_profile ?")
    if answer:
        with open(Path("~/.bash_profile").expanduser(), 'a') as fd:
            fd.write(
                textwrap.dedent(
                    f"""
                    # Environment for project {project} added by `cgse init` at {format_datetime()}
                    export PROJECT={project}
                    export SITE_ID={site_id}
                    export {project}_DATA_STORAGE_LOCATION={data_storage_location}
                    export {project}_CONF_DATA_LOCATION={conf_data_location}
                    export {project}_LOG_FILE_LOCATION={log_file_location}
                    export {project}_LOCAL_SETTINGS={local_settings_path}
                    """
                )
            )
    else:
        rich.print(
            textwrap.dedent(
                f"""
                # -> Add the following lines to your bash profile or equivalent
                
                export PROJECT={project}
                export SITE_ID={site_id}
                export {project}_DATA_STORAGE_LOCATION={data_storage_location}
                export {project}_CONF_DATA_LOCATION={conf_data_location}
                export {project}_LOG_FILE_LOCATION={log_file_location}
                export {project}_LOCAL_SETTINGS={local_settings_path}
            """
            )

        )


show = typer.Typer(help="Show information about settings, environment, setup, ...", no_args_is_help=True)


@show.command(name="settings")
def show_settings():
    """Show the settings that are defined by the installed packages."""
    proc = subprocess.Popen(
        [sys.executable, "-m", "egse.settings"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    stdout, stderr = proc.communicate()
    rich.print(stdout.decode(), end='')
    if stderr:
        rich.print(f"[red]{stderr.decode()}[/]")


@show.command(name="env")
def show_env(
        mkdir: Annotated[bool, typer.Option(help="Create the missing folder")] = None,
        full: Annotated[bool, typer.Option(help="Provide additional info")] = None,
        doc: Annotated[bool, typer.Option(help="Provide documentation on environment variables")] = None,
):
    """Show the environment variables that are defined for the project."""
    options = [opt for opt, flag in [("--mkdir", mkdir), ("--full", full), ("--doc", doc)] if flag]

    cmd = [sys.executable, "-m", "egse.env"]
    cmd += options if options else []

    proc = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    stdout, stderr = proc.communicate()
    rich.print(stdout.decode(), end='')
    if stderr:
        rich.print(f"[red]{stderr.decode()}[/]")


@show.command(name="procs")
def show_processes():
    """Show the processes that are running for the installed packages."""
    for ep in entry_points("cgse.explore"):
        # print(f"{ep.name = }, {ep.module = }, {ep.load() = }, {ep.extras = }")
        explore = ep.load()
        with contextlib.suppress(AttributeError):
            show_procs = getattr(explore, "show_processes")
            for line in show_procs():
                rich.print(line)


check = typer.Typer(help="Check installation, settings, required files, etc.", no_args_is_help=True)


@check.command(name="setups")
def check_setups(verbose: bool = False):
    """Perform a number of checks on the SETUP files."""

    logging.basicConfig(
        level=logging.CRITICAL,
        format="[%(asctime)s] %(threadName)-12s %(levelname)-8s %(name)-20s %(lineno)5d:%(module)-20s %(message)s",
    )

    # What can we check with respect to the setups?
    #
    # - CONF_DATA_LOCATION

    from egse.env import get_conf_data_location
    from egse.env import get_site_id
    from egse.config import find_files

    any_errors = 0
    error_messages = []

    try:
        conf_data_location = get_conf_data_location()
    except ValueError:
        conf_data_location = None

    site_id = get_site_id()

    # ---------- check if the Site_ID is set

    verbose and rich.print("Checking site id ...")

    if not site_id:
        any_errors += 1
        error_messages.append(
            "[red]The environment variable SITE_ID is not set. "
            "SITE_ID is required to define the project settings and environment variables.[/]"
        )

    # ---------- check if the <PROJECT>_CONF_DATA_LOCATION is set

    verbose and rich.print("Checking configuration data location ...")

    if not conf_data_location:
        any_errors += 1
        error_messages.append(
            "[red]The location of the configuration data can not be determined, check your environment using `cgse "
            "show env`.[/]"
        )
    elif not Path(conf_data_location).exists():
        any_errors += 1
        error_messages.append(
            f"[red]The location of the configuration data doesn't exist: {conf_data_location!s}[/]"
        )

    if any_errors:
        print_error_messages(error_messages)
        return

    # ---------- check if there is at least one SETUP in the configuration data folder

    verbose and rich.print("Checking available SETUP files ...")

    files = list(find_files("SETUP*.yaml", root=conf_data_location))

    if not files:
        any_errors += 1
        error_messages.append(
            f"[red]No SETUP files were found at {conf_data_location}[/]"
        )
    else:
        verbose and rich.print(f":arrow_forward: Found {len(files)} Setup files.")

    regex = re.compile(f"SETUP_{site_id}_00000_.*.yaml")

    if not any(True for file in files if regex.search(str(file))):
        any_errors += 1
        error_messages.append(
            f"[red]There is no 'Zero' SETUP for {site_id} in {conf_data_location}[/]"
        )
    else:
        verbose and rich.print(f":arrow_forward: Found the 'Zero' Setup file: SETUP_{site_id}_00000_*.yaml")

    if any_errors:
        print_error_messages(error_messages)
        return

    # ---------- check for each SETUP file if the site_id matches the SITE_ID

    verbose and rich.print("Checking site_id in setup files ...")

    for file in files:
        setup = Setup.from_yaml_file(file)
        this_site_id = setup.get("site_id")
        if this_site_id is None:
            any_errors += 1
            error_messages.append(f"There is no 'site_id' defined in '{file!s}'")
        elif this_site_id != site_id:
            any_errors = 1
            error_messages.append(
                f"[red]The site_id ('{this_site_id}') in '{file!s}' doesn't match the environment "
                f"variable SITE_ID={site_id}[/]"
            )

    if any_errors:
        print_error_messages(error_messages)
    else:
        rich.print("[green]everything seems to be ok.[/]")


def print_error_messages(messages: list[str]):
    from rich.panel import Panel

    content = "\n"

    for msg in messages:
        content += f":arrow_forward: {msg}\n"
    content += "\n[orange1]Fix the above errors before proceeding.[/]"

    rich.print(Panel(content, title="Errors found during analysis"))



"""
This module provides convenience functions to inspect and compare dictionaries when debugging.
"""

__all__ = [
    "flatten_dict",
    "log_differences",
    "log_key_differences",
]
import logging

from egse.system import capture_rich_output
from egse.system import flatten_dict
from rich.table import Table

LOGGER = logging.getLogger(__name__)


def log_differences(dict_1, dict_2):
    """
    Takes two flattened dictionaries and compares them. This function only compares those
    keys that are common to both dictionaries. The key-value pairs that are unique to one
    of the dictionaries are ignored. To inspect if there are keys unique to one dictionary,
    use the [log_key_differences()](dicts.md#egse.dicts.log_key_differences) function.

    The differences are logged in a Rich Table at level=INFO.

    Example:
        ```text
        >>> d1 = { "A": 1, "B": 2, "C": 3 }
        >>> d2 = { "A": 1, "B": 5, "C": 3 }
        >>> log_differences(d1, d2)

        2025-02-28 09:20:51,639:         MainProcess:    INFO:   37:__main__            :Value Differences

        ┏━━━━━━┳━━━━━━━━━━━┳━━━━━━━━━━━┓
        ┃ Name ┃ old value ┃ new value ┃
        ┡━━━━━━╇━━━━━━━━━━━╇━━━━━━━━━━━┩
        │ B    │ 2         │ 5         │
        └──────┴───────────┴───────────┘
        ```
    """

    all_keys = dict_1.keys() & dict_2.keys()

    if mismatched := {key for key in all_keys if dict_1[key] != dict_2[key]}:
        table = Table("Name", "old value", "new value", title="Value Differences", title_justify="left")

        for name in sorted(mismatched):
            table.add_row(name, str(dict_1[name]), str(dict_2[name]))

        LOGGER.info(capture_rich_output(table))
        # rich.print(table)
    else:
        LOGGER.info(f"No differences between the two flattened dictionaries, {len(all_keys)} values compared.")


def log_key_differences(dict_1, dict_2):
    """
    Takes two dictionaries and compares the top-level keys. The differences are logged in a Rich Table at level=INFO.
    Keys that are present on both dictionaries are not logged.

    Example:
        ```text
        >>> d1 = {"A": 1, "B": 2, "C": 3}
        >>> d2 = {"B": 2, "C": 3, "D": 4}
        >>> log_key_differences(d1, d2)
        2025-02-28 09:08:29,916:         MainProcess:    INFO:   60:__main__            :Key differences
        ┏━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━┓
        ┃ Dictionary 1 ┃ Dictionary 2 ┃
        ┡━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━┩
        │ A            │              │
        │              │ D            │
        └──────────────┴──────────────┘
        ```
    """
    s1 = set(dict_1)
    s2 = set(dict_2)

    not_in_s2 = s1 - s2
    not_in_s1 = s2 - s1

    if not not_in_s1 and not not_in_s2:
        LOGGER.info("Both dictionaries contains the same keys.")

    table = Table("Dictionary 1", "Dictionary 2", title="Key differences", title_justify="left")

    for key in not_in_s2:
        table.add_row(str(key), "")

    for key in not_in_s1:
        table.add_row("", str(key))

    LOGGER.info(capture_rich_output(table))


if __name__ == "__main__":
    import egse.logger  # noqa activates the logger

    d1 = {
        "A": 1,
        "B": 2,
        "C": 3,
    }

    d2 = {
        "B": 2,
        "C": 3,
        "D": 4,
    }

    d3 = {
        "A": 1,
        "B": 5,
        "C": 3,
    }

    log_differences(d1, d2)
    log_key_differences(d1, d2)

    log_differences(d1, d3)



import asyncio
import functools
import time

import pytest

from egse.decorators import execution_count
from egse.task import AwaitTask
from egse.task import task


@execution_count
def plain_func(msg: str, sleep: float = 1.0):
    time.sleep(sleep)
    return f"plain_func: I got this message: {msg}, waited {sleep}s before executing"


@execution_count
async def async_func(msg: str, sleep: float = 1.0):

    await asyncio.sleep(sleep)
    return f"async_func: I got this message: {msg}, waited {sleep}s before executing"


def test_synchronous():

    plain_func.reset()

    t = task(plain_func, "Hello, plain old Python function", 0.0)
    assert isinstance(t, AwaitTask)
    assert "plain" in t()
    assert "plain" in t.execute()
    assert plain_func.counts() == 2

    async_func.reset()

    t = task(async_func, "hello, async function", 0.0)
    assert isinstance(t, AwaitTask)
    assert "async" in t()
    assert "async" in t.execute()
    assert async_func.counts() == 2


@pytest.mark.asyncio
async def test_asynchronous():

    plain_func.reset()

    t = task(plain_func, "Hello, plain old Python function", 0.0)
    assert isinstance(t, AwaitTask)
    assert "plain" in await t
    assert "plain" in t()
    assert "plain" in t.execute()
    assert plain_func.counts() == 3

    async_func.reset()

    t = task(async_func, "Hello, async function", 0.0)
    assert isinstance(t, AwaitTask)
    assert "async" in await t
    assert "async" in await t()
    assert "async" in await t.execute()
    assert async_func.counts() == 3



"""
The Log Server receives all log messages and events from control servers and client applications
and saves those messages in a log file at a given location. The log messages are retrieved over
a ZeroMQ message channel.
"""
__all__ = []

import datetime
import logging
import multiprocessing
import pickle
import sys
from logging import StreamHandler
from logging.handlers import SocketHandler
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path
from typing import Optional

import rich
import typer
import zmq

from egse.env import get_log_file_location
from egse.logger import LOGGER_ID
from egse.logger import get_log_file_name
from egse.logger import send_request
from egse.process import SubProcess
from egse.registry.client import RegistryClient
from egse.settings import Settings
from egse.signal import FileBasedSignaling
from egse.system import format_datetime
from egse.system import get_caller_info
from egse.system import get_host_ip
from egse.zmq_ser import bind_address
from egse.zmq_ser import get_port_number

CTRL_SETTINGS = Settings.load("Logging Control Server")

LOG_NAME_TO_LEVEL = {
    'CRITICAL': 50,
    'FATAL': 50,
    'ERROR': 40,
    'WARN': 30,
    'WARNING': 30,
    'INFO': 20,
    'DEBUG': 10,
    'NOTSET': 0,
}

# The format for the log file.
# The line that is saved in the log file shall contain as much information as possible.

LOG_FORMAT_FILE = (
    "%(asctime)s:%(processName)s:%(process)s:%(levelname)s:%(lineno)d:%(name)s:%(message)s"
)

LOG_FORMAT_KEY_VALUE = (
    "level=%(levelname)s ts=%(asctime)s process=%(processName)s process_id=%(process)s "
    "name=%(name)s caller=%(filename)s:%(lineno)s function=%(funcName)s msg=\"%(message)s\""
)

LOG_FORMAT_DATE = "%Y-%m-%dT%H:%M:%S,%f"

# The format for the console output.
# The line that is printed on the console shall be concise.

LOG_FORMAT_STREAM = "%(asctime)s:%(levelname)s:%(name)s:%(filename)s:%(funcName)s:%(message)s"

LOG_LEVEL_FILE = logging.DEBUG
LOG_LEVEL_STREAM = logging.ERROR
LOG_LEVEL_SOCKET = 1  # ALL records shall go to the socket handler

LOGGER_NAME = "egse.logger"

PROTOCOL = 'tcp'
RECEIVER_PORT = 0  # dynamically assigned by the system
COMMANDER_PORT = 0  # dynamically assigned by the system

file_handler: Optional[TimedRotatingFileHandler] = None
stream_handler: Optional[StreamHandler] = None
socket_handler: Optional[SocketHandler] = None


class DateTimeFormatter(logging.Formatter):

    def formatTime(self, record, datefmt=None):
        converted_time = datetime.datetime.fromtimestamp(record.created)
        if datefmt:
            return converted_time.strftime(datefmt)
        formatted_time = converted_time.strftime("%Y-%m-%dT%H:%M:%S")
        return f"{formatted_time}.{record.msecs:03.0f}"


file_formatter = DateTimeFormatter(fmt=LOG_FORMAT_KEY_VALUE, datefmt=LOG_FORMAT_DATE)

app_name = "log_cs"
app = typer.Typer(name=app_name, no_args_is_help=True)


@app.command()
def start():
    """Start the Logger Control Server."""

    global file_handler, stream_handler, socket_handler

    multiprocessing.current_process().name = app_name

    log_file_location = Path(get_log_file_location())
    log_file_name = get_log_file_name()

    if not log_file_location.exists():
        raise FileNotFoundError(f"The location for the log files doesn't exist: {log_file_location!s}.")

    file_handler = TimedRotatingFileHandler(filename=log_file_location / log_file_name, when='midnight')
    file_handler.setFormatter(file_formatter)

    # There is no need to set the level for the handlers, because the level is checked by the
    # Logger, and we use the handlers directly here. Use a filter to restrict messages.

    stream_handler = StreamHandler()
    stream_handler.setFormatter(logging.Formatter(fmt=LOG_FORMAT_STREAM))

    # Log records are also sent to the textualog listening server

    socket_handler = SocketHandler(CTRL_SETTINGS.TEXTUALOG_IP_ADDRESS, CTRL_SETTINGS.TEXTUALOG_LISTENING_PORT)
    socket_handler.setFormatter(file_formatter)

    context = zmq.Context.instance()

    endpoint = bind_address(PROTOCOL, RECEIVER_PORT)
    receiver = context.socket(zmq.PULL)
    receiver.bind(endpoint)

    endpoint = bind_address(PROTOCOL, COMMANDER_PORT)
    commander = context.socket(zmq.REP)
    commander.bind(endpoint)

    poller = zmq.Poller()
    poller.register(receiver, zmq.POLLIN)
    poller.register(commander, zmq.POLLIN)

    client = RegistryClient()
    client.connect()
    service_id = client.register(
        name=LOGGER_ID,
        host=get_host_ip() or "127.0.0.1",
        port=get_port_number(commander),
        service_type="LOGGER",
        metadata={
            'receiver_port': get_port_number(receiver),
        }
    )
    if service_id is None:
        record = _create_log_record(logging.ERROR, "Registration of LOGGER service failed.")
        handle_log_record(record)
        return

    client.start_heartbeat()

    def reregister_service(force: bool = False):
        nonlocal service_id

        record = _create_log_record(logging.WARNING, f"Re-registration of Logger {force = }.")
        handle_log_record(record)

        if client.get_service(service_id):
            if force is True:
                client.deregister(service_id)
            else:
                return

        service_id = client.register(
            name=LOGGER_ID,
            host=get_host_ip() or "127.0.0.1",
            port=get_port_number(commander),
            service_type="LOGGER",
            metadata={
                'receiver_port': get_port_number(receiver),
            }
        )
        if service_id is None:
            record = _create_log_record(logging.ERROR, "Registration of LOGGER service failed.")
            handle_log_record(record)

    signaling = FileBasedSignaling(app_name)
    signaling.start_monitoring()
    signaling.register_handler("reregister", reregister_service)

    while True:
        try:
            signaling.process_pending_commands()

            socks = dict(poller.poll(timeout=1000))  # timeout in milliseconds

            if commander in socks:
                pickle_string = commander.recv()
                command = pickle.loads(pickle_string)

                if command.lower() == "quit":
                    commander.send(pickle.dumps("ACK"))
                    break

                response = handle_command(command)
                commander.send(pickle.dumps(response))

            if receiver in socks:
                pickle_string = receiver.recv()
                record = pickle.loads(pickle_string)
                record = logging.makeLogRecord(record)
                handle_log_record(record)

        except KeyboardInterrupt:
            rich.print("KeyboardInterrupt caught!")
            break

    record = _create_log_record(level=logging.WARNING, msg="Logger terminated.")
    handle_log_record(record)

    file_handler.close()
    stream_handler.close()
    commander.close(linger=0)
    receiver.close(linger=0)

    client.stop_heartbeat()
    client.deregister(service_id)
    client.disconnect()


def _create_log_record(level: int, msg: str) -> logging.LogRecord:
    """Create a LogRecord that can be handled by a Handler."""
    caller_info = get_caller_info(level=2)

    record = logging.LogRecord(
        name=LOGGER_NAME,
        level=level,
        pathname=caller_info.filename,
        lineno=caller_info.lineno,
        msg=msg,
        args=(),
        exc_info=None,
        func=caller_info.function,
        sinfo=None
    )

    return record


@app.command()
def start_bg():
    """Start the Logger Control Server in the background."""
    proc = SubProcess("log_cs", ["log_cs", "start"])
    proc.execute()


def handle_log_record(record):
    """Send the log record to the file handler and the stream handler."""
    global file_handler, stream_handler, socket_handler

    if record.levelno >= LOG_LEVEL_FILE:
        file_handler.emit(record)

    if record.levelno >= LOG_LEVEL_STREAM:
        stream_handler.handle(record)

    if record.levelno >= LOG_LEVEL_SOCKET:
        socket_handler.handle(record)


def handle_command(command) -> dict:
    """Handle commands that are sent to the commanding socket."""
    global file_handler
    global LOG_LEVEL_FILE

    response = dict(
        timestamp=format_datetime(),
    )
    if command.lower() == 'roll':
        file_handler.doRollover()
        response.update(dict(status="ACK"))
        record = logging.LogRecord(
            name=LOGGER_NAME,
            level=logging.WARNING,
            pathname=__file__,
            lineno=197,
            msg="Logger rolled over.",
            args=(),
            exc_info=None,
            func="roll",
            sinfo=None
        )
        handle_log_record(record)

    elif command.lower() == 'status':
        with RegistryClient() as client:
            service = client.discover_service("LOGGER")
        if service:
            response.update(dict(
                status="ACK",
                logging_port=service['metadata']['receiver_port'],
                commanding_port=service['port'],
                file_logger_level=logging.getLevelName(LOG_LEVEL_FILE),
                stream_logger_level=logging.getLevelName(LOG_LEVEL_STREAM),
                file_logger_location=file_handler.baseFilename,
            ))
        else:
            response.update(
                dict(status="NACK")
            )
    elif command.lower().startswith("set_level"):
        new_level = command.split()[-1]
        LOG_LEVEL_FILE = LOG_NAME_TO_LEVEL[new_level]
        response.update(dict(
            status="ACK",
            file_logger_level=logging.getLevelName(LOG_LEVEL_FILE),
        ))

    return response


@app.command()
def stop():
    """Stop the Logger Control Server."""

    response = send_request("quit")
    if response == "ACK":
        rich.print("Logger successfully terminated.")
    else:
        rich.print(f"[red] ERROR: {response}")


@app.command()
def roll():
    """Roll over the log file of the Logger Control Server."""

    response = send_request("roll")
    if response.get("status") == "ACK":
        rich.print("[green]Logger files successfully rotated.")
    else:
        rich.print(f"[red]ERROR: {response}")


@app.command()
def status():
    """Roll over the log file of the Logger Control Server."""

    response = send_request("status")
    if response.get("status") == "ACK":
        rich.print("Log Manager:")
        rich.print("    Status: [green]active")
        rich.print(f"    Logging port: {response.get('logging_port')}")
        rich.print(f"    Commanding port: {response.get('commanding_port')}")
        rich.print(f"    Level [grey50](file)[black]: {response.get('file_logger_level')}")
        rich.print(f"    Level [grey50](stdout)[black]: {response.get('stream_logger_level')}")
        rich.print(f"    Log file location: {response.get('file_logger_location')}")
    else:
        rich.print("Log Manager Status: [red]not active")


@app.command()
def test():
    # setup_logging() and teardown_logging() is automatic
    # setup_logging()

    logger = logging.getLogger('egse')
    logger.debug("A DEBUG message")
    logger.info("An INFO message")
    logger.warning("A WARNING message")

    # from egse.logger import print_all_handlers
    # print_all_handlers()

    # teardown_logging()


@app.command()
def set_level(level: str):
    """Set the logging level for """
    try:
        level = logging.getLevelName(int(level))
    except ValueError:
        if level not in LOG_NAME_TO_LEVEL:
            rich.print(f"[red]Invalid logging level given '{level}'.")
            rich.print(f"Should be one of: {', '.join(LOG_NAME_TO_LEVEL.keys())}.")
            return

    response = send_request(f"set_level {level}")
    if response.get("status") == "ACK":
        rich.print(f"Log level on the server is now set to {response.get('file_logger_level')}.")
    else:
        rich.print(f"[red]ERROR: {response}")


if __name__ == "__main__":

    logging.basicConfig(level=logging.INFO)

    sys.exit(app())



import asyncio
import json
import time

import pytest

from egse.device import DeviceConnectionError
from egse.device import DeviceTimeoutError
from egse.tempcontrol.keithley.daq6510_adev import DAQ6510

SCPI_PORT = 5025
HOSTNAME = "192.168.68.77"  # hostname at home with DHCP


@pytest.mark.sayncio
async def is_daq6510_available():

    try:
        daq = DAQ6510(HOSTNAME, SCPI_PORT)
        await daq.connect()
    except DeviceTimeoutError:
        return False

    return True


@pytest.mark.asyncio
async def test_constructor():

    daq = DAQ6510(HOSTNAME, SCPI_PORT)
    await daq.connect()

    response = await daq.query("*IDN?\n")

    assert response == "KEITHLEY INSTRUMENTS,MODEL DAQ6510,04569510,1.7.12b"

    await daq.disconnect()


@pytest.mark.asyncio
async def test_connection(caplog):

    daq = DAQ6510(HOSTNAME, SCPI_PORT)

    assert not await daq.is_connected()

    await daq.connect()

    assert await daq.is_connected()

    await daq.disconnect()

    assert not await daq.is_connected()

    await daq.disconnect()  # This should do nothing

    await daq.connect()

    caplog.clear()

    await daq.connect()  # calling this when the connection is open issues a warning

    assert "Trying to connect to an already connected device" in caplog.text

    await daq.disconnect()

    assert not await daq.is_connected()

    await daq.reconnect()
    assert await daq.is_connected()

    await daq.reconnect()
    assert await daq.is_connected()

    await daq.disconnect()


@pytest.mark.asyncio
async def test_context_manager():

    async with DAQ6510(HOSTNAME, SCPI_PORT) as daq:
        assert await daq.is_connected()

    assert not await daq.is_connected()


@pytest.mark.asyncio
async def test_incorrect_construction():

    daq = DAQ6510(HOSTNAME)
    daq.hostname = "unknown"  # set this explicitly because the local_settings might define the HOSTNAME
    with pytest.raises(DeviceConnectionError, match="DAQ6510: Address resolution error for unknown"):
        await daq.connect()

    daq = DAQ6510(HOSTNAME)
    daq.hostname = None
    with pytest.raises(ValueError, match="DAQ6510: Hostname is not initialized"):
        await daq.connect()

    daq = DAQ6510(HOSTNAME)
    daq.port = None

    with pytest.raises(ValueError, match="DAQ6510: Port number is not initialized"):
        await daq.connect()

    daq = DAQ6510(HOSTNAME, 3000)  # pass an incorrect port number

    with pytest.raises(DeviceConnectionError, match="DAQ6510: Connection refused"):
        await daq.connect()


@pytest.mark.asyncio
async def test_write_read():

    async with DAQ6510(HOSTNAME, SCPI_PORT) as daq:

        await daq.write(":SYST:COMM:LAN:CONF \"AUTO\"")
        await daq.write(":SYST:COMM:LAN:CONF?")

        response = (await daq.read()).decode()

        assert response.startswith("auto")


@pytest.mark.asyncio
async def test_a_scan():

    async with DAQ6510(HOSTNAME, SCPI_PORT) as daq:

        # Initialize

        await daq.write('*RST')  # this also the user-defined buffer "test1"

        for cmd, response in [
            ('TRAC:MAKE "test1", 1000', False),  # create a new buffer
            # settings for channel 1 and 2 of slot 1
            ('SENS:FUNC "TEMP", (@101:102)', False),  # set the function to temperature

            ("SENS:TEMP:TRAN FRTD, (@101)", False),  # set the transducer to 4-wire RTD
            ("SENS:TEMP:RTD:FOUR PT100, (@101)", False),  # set the type of the 4-wire RTD

            ("SENS:TEMP:TRAN RTD, (@102)", False),  # set the transducer to 2-wire RTD
            ("SENS:TEMP:RTD:TWO PT100, (@102)", False),  # set the type of the 2-wire RTD

            ('ROUT:SCAN:BUFF "test1"', False),
            ("ROUT:SCAN:CRE (@101:102)", False),
            ("ROUT:CHAN:OPEN (@101:102)", False),
            ("ROUT:STAT? (@101:102)", True),
            ("ROUT:SCAN:STAR:STIM NONE", False),
            # ("ROUT:SCAN:ADD:SING (@101, 102)", False),  # not sure what this does, not really needed
            ("ROUT:SCAN:COUN:SCAN 1", False),  # not sure if this is needed in this setting
            # ("ROUT:SCAN:INT 1", False),
        ]:
            if response:
                print(f"Sending {cmd}... response=", end="")
                print(await daq.trans(cmd))
            else:
                print(f"Sending {cmd}...")
                await daq.write(cmd)

        # Read out the channels

        # daq.write('TRAC:CLE "test1"\n')

        for _ in range(10):
            await daq.write("INIT:IMM")
            await daq.write("*WAI")

            # Reading the data

            # When a trigger mode is running, these READ? commands can not be used.

            # print(daq.trans('READ? "test1", CHAN, TST, READ\n', wait=False), end="")
            # print(daq.trans('READ? "test1", CHAN, TST, READ\n', wait=False), end="")
            # time.sleep(1)
            # print(daq.trans('READ? "test1", CHAN, TST, READ\n', wait=False), end="")
            # print(daq.trans('READ? "test1", CHAN, TST, READ\n', wait=False), end="")

            # Read out the buffer

            response = await daq.trans('TRAC:DATA? 1, 2, "test1", CHAN, TST, READ')
            print(f"{response = }")
            ch1, tst1, val1, ch2, tst2, val2 = response.split(",")
            print(f"Channel: {ch1} Time: {tst1} Value: {float(val1):.4f}\t", end="")
            print(f"Channel: {ch2} Time: {tst2} Value: {float(val2):.4f}")
            await asyncio.sleep(0.1)


@pytest.mark.asyncio
async def test_another_scan():

    async with DAQ6510(HOSTNAME, SCPI_PORT) as daq:

        # Initialize

        n_readings = 50

        await daq.write('*RST')  # This also clears the default buffers

        for cmd, response in [
            ('SENS:FUNC "TEMP", (@101,102)', False),
            ('SENS:TEMP:UNIT CELS, (@101,102)', False),

            ("SENS:TEMP:TRAN FRTD, (@101)", False),  # set the transducer to 4-wire RTD
            ("SENS:TEMP:RTD:FOUR PT100, (@101)", False),  # set the type of the 4-wire RTD

            ("SENS:TEMP:TRAN RTD, (@102)", False),  # set the transducer to 2-wire RTD
            ("SENS:TEMP:RTD:TWO PT100, (@102)", False),  # set the type of the 2-wire RTD

            # Set the amount of time that the input signal is measured.
            ('SENS:TEMP:NPLC 0.1, (@101,102)', False),

            ('TRIG:LOAD "Empty"', False),
            ('TRIG:BLOC:BUFF:CLEAR 1, "defbuffer1"', False),
            ('TRIG:BLOC:MDIG 2, "defbuffer1"', False),
            (f'TRIG:BLOC:BRAN:COUN 3, {n_readings}, 2', False),
        ]:
            if response:
                print(f"Sending {cmd}... response=", end="")
                print(await daq.trans(cmd))
            else:
                print(f"Sending {cmd}...")
                await daq.write(cmd)

        # Read out the channels

        # What should this read_timeout be and how shall I set and reset this?
        daq.read_timeout = 10.0

        for _ in range(10):
            await daq.write("INIT:IMM")
            await daq.write("*WAI")

            # Reading the data

            response = await daq.trans(f'TRAC:DATA? 1, {n_readings}, "defbuffer1", CHAN, TST, READ')
            print(f"{response = }")
            # ch1, tst1, val1, ch2, tst2, val2 = response.split(",")
            # print(f"Channel: {ch1} Time: {tst1} Value: {float(val1):.4f}\t", end="")
            # print(f"Channel: {ch2} Time: {tst2} Value: {float(val2):.4f}")
            await asyncio.sleep(0.1)


@pytest.mark.asyncio
async def test_daq6510_mon():

    from egse.tempcontrol.keithley.daq6510_mon import DAQMonitorClient

    client = DAQMonitorClient(server_address="localhost", port=5556)

    try:
        # Get current status
        status = client.get_status()
        print(f"Service status: {json.dumps(status, indent=2)}")

        # Start polling with custom settings
        response = client.start_polling(channels=["101", "102"], interval=2.0)
        print(f"Start polling response: {response}")

        # Wait for a while
        time.sleep(5)

        # Change polling interval
        response = client.set_interval(3.0)
        print(f"Set interval response: {response}")

        # Wait for a while
        time.sleep(5)

        # Stop polling
        response = client.stop_polling()
        print(f"Stop polling response: {response}")

    finally:
        client.close()



import logging
from enum import Enum

# Default ports that are assigned to REQ-REP and PUB-SUB protocols of the registry services
DEFAULT_RS_REQ_PORT = 4242  # Handle requests
DEFAULT_RS_PUB_PORT = 4243  # Publish events
DEFAULT_RS_HB_PORT = 4244  # Heartbeats

DEFAULT_RS_DB_PATH = 'service_registry.db'


class MessageType(Enum):
    """Message types using the envelope frame in the ROUTER-DEALER protocol."""
    REQUEST_WITH_REPLY = b"REQ"     # Client expects a reply
    REQUEST_NO_REPLY = b"NO-REQ"    # No reply expected by the client
    RESPONSE = b"REP"               # Response to a request
    NOTIFICATION = b"NOTIF"         # Server-initiated notification
    HEARTBEAT = b"HB"               # Heartbeat/health check


logger = logging.getLogger("egse.registry")



import pickle

from egse.zmq_ser import zmq_error_response
from egse.zmq_ser import zmq_json_request
from egse.zmq_ser import zmq_json_response
from egse.zmq_ser import zmq_string_request
from egse.zmq_ser import zmq_string_response


def test_zmq_requests():

    assert zmq_string_request("") == [b'MESSAGE_TYPE:STRING', b'']
    assert zmq_string_request("hello") == [b"MESSAGE_TYPE:STRING", b"hello"]

    assert zmq_json_request({}) == [b'MESSAGE_TYPE:JSON', b'{}']
    assert zmq_json_request({'command': 'do_something'}) == [b"MESSAGE_TYPE:JSON", b'{"command": "do_something"}']


def test_zmq_responses():

    assert zmq_string_response("") == [b'MESSAGE_TYPE:STRING', b'']
    assert zmq_string_response("Good Job!") == [b'MESSAGE_TYPE:STRING', b'Good Job!']

    assert zmq_json_response({}) == [b'MESSAGE_TYPE:JSON', b'{}']
    assert zmq_json_response({'success': True, 'message': "all good", 'metadata': {'version': '0.1.0'}}) == [
        b'MESSAGE_TYPE:JSON',
        b'{"success": true, "message": "all good", "metadata": {"version": "0.1.0"}}'
    ]
    assert zmq_json_response({'success': True,
                              'message': {'name': "cgse", 'description': "Great software"},
                              'metadata': {'version': '0.1.0'}}) == [
        b'MESSAGE_TYPE:JSON',
        b'{'
        b'"success": true, '
        b'"message": {"name": "cgse", "description": "Great software"}, "metadata": {"version": "0.1.0"}'
        b'}'
    ]


def test_zmq_error_responses():

    assert zmq_error_response({}) == [b'MESSAGE_TYPE:ERROR', b'\x80\x04}\x94.']
    response = zmq_error_response(
        {
            'success': False,
            'message': "Incorrect message type: MESSAGE_TYPE:PICKLE",
            'metadata': {
                'data': {
                    'file': '/Users/rik/github/cgse/libs/cgse-common/src/egse/async_control.py',
                    'lineno': 322,
                    'function': '_process_device_command'
                },
            },
        }
    )
    assert isinstance(response, list)
    assert len(response) == 2
    assert response[0] == b'MESSAGE_TYPE:ERROR'
    assert isinstance(response[1], bytes)
    assert pickle.loads(response[1]) == {
            'success': False,
            'message': "Incorrect message type: MESSAGE_TYPE:PICKLE",
            'metadata': {
                'data': {
                    'file': '/Users/rik/github/cgse/libs/cgse-common/src/egse/async_control.py',
                    'lineno': 322,
                    'function': '_process_device_command'
                },
            },
        }



"""
The Control Server that connects to the Hexapod ZONDA Hardware Controller.

Start the control server from the terminal as follows:

    $ zonda_cs start-bg

or when you don't have the device available, start the control server in simulator mode. That
will make the control server connect to a device software simulator:

    $ zonda_cs start --sim

Please note that software simulators are intended for simple test purposes and will not simulate
all device behavior correctly, e.g. timing, error conditions, etc.

"""
import multiprocessing
import sys
from typing import Annotated

import rich
import typer
import zmq
from prometheus_client import start_http_server

from egse.control import ControlServer
from egse.control import is_control_server_active
from egse.hexapod.symetrie import ProxyFactory
from egse.hexapod.symetrie import get_hexapod_controller_pars
from egse.hexapod.symetrie import logger
from egse.hexapod.symetrie.zonda_protocol import ZondaProtocol
from egse.registry.client import RegistryClient
from egse.services import ServiceProxy
from egse.settings import Settings
from egse.storage import store_housekeeping_information
from egse.zmq_ser import connect_address

CTRL_SETTINGS = Settings.load("Hexapod Control Server")


class ZondaControlServer(ControlServer):
    """ZondaControlServer - Command and monitor the Hexapod ZONDA hardware.

    This class works as a command and monitoring server to control the Symétrie Hexapod PUNA.
    This control server shall be used as the single point access for controlling the hardware
    device. Monitoring access should be done preferably through this control server also,
    but can be done with a direct connection through the PunaController if needed.

    The sever binds to the following ZeroMQ sockets:

    * a REQ-REP socket that can be used as a command server. Any client can connect and
      send a command to the Hexapod.

    * a PUB-SUP socket that serves as a monitoring server. It will send out Hexapod status
      information to all the connected clients every five seconds.

    """

    def __init__(self, device_id: str, simulator: bool = False):
        super().__init__()

        multiprocessing.current_process().name = "zonda_cs"

        self.logger = logger

        self.device_id = device_id
        self.device_protocol = ZondaProtocol(self, device_id=device_id, simulator=simulator)

        self.device_protocol.bind(self.dev_ctrl_cmd_sock)

        self.poller.register(self.dev_ctrl_cmd_sock, zmq.POLLIN)

        self.register_service(service_type=f"{device_id}")

    def get_communication_protocol(self):
        return CTRL_SETTINGS.PROTOCOL

    def get_commanding_port(self):
        return CTRL_SETTINGS.COMMANDING_PORT

    def get_service_port(self):
        return CTRL_SETTINGS.SERVICE_PORT

    def get_monitoring_port(self):
        return CTRL_SETTINGS.MONITORING_PORT

    def get_storage_mnemonic(self):
        try:
            return CTRL_SETTINGS.STORAGE_MNEMONIC
        except AttributeError:
            return "ZONDA"

    def is_storage_manager_active(self):
        from egse.storage import is_storage_manager_active
        return is_storage_manager_active()

    def store_housekeeping_information(self, data):
        """Send housekeeping information to the Storage manager."""

        origin = self.get_storage_mnemonic()
        store_housekeeping_information(origin, data)

    def register_to_storage_manager(self):
        from egse.storage import register_to_storage_manager
        from egse.storage.persistence import TYPES

        register_to_storage_manager(
            origin=self.get_storage_mnemonic(),
            persistence_class=TYPES["CSV"],
            prep={
                "column_names": list(self.device_protocol.get_housekeeping().keys()),
                "mode": "a",
            }
        )

    def unregister_from_storage_manager(self):
        from egse.storage import unregister_from_storage_manager

        unregister_from_storage_manager(origin=self.get_storage_mnemonic())

    def before_serve(self):
        start_http_server(CTRL_SETTINGS.METRICS_PORT)

    def after_serve(self) -> None:
        self.deregister_service()


app = typer.Typer()


@app.command()
def start(
        device_id: Annotated[
            str,
            typer.Argument(help="the device identifier, identifies the hardware controller")
        ],
        simulator: Annotated[
            bool,
            typer.Option("--simulator", "--sim", help="start the hexapod PUNA Control Server in simulator mode")
        ] = False
):
    """Start the Hexapod Zonda Control Server."""

    try:

        controller = ZondaControlServer(device_id, simulator)
        controller.serve()

    except KeyboardInterrupt:

        print("Shutdown requested...exiting")

    except SystemExit as exc:

        exit_code = exc.code if hasattr(exc, 'code') else 0
        print(f"System Exit with code {exc.code}")
        sys.exit(exit_code)

    except Exception:

        logger.exception("Cannot start the Hexapod Zonda Control Server")

        # The above line does exactly the same as the traceback, but on the logger
        # import traceback
        # traceback.print_exc(file=sys.stdout)

    return 0


@app.command()
def stop(device_id: str):
    """Send a 'quit_server' command to the Hexapod Zonda Control Server."""

    with RegistryClient() as reg:
        service = reg.discover_service(device_id)
        rich.print("service = ", service)

        if service:
            proxy = ServiceProxy(protocol="tcp", hostname=service["host"], port=service['metadata']['service_port'])
            proxy.quit_server()
        else:
            *_, device_type, controller_type = get_hexapod_controller_pars(device_id)

            factory = ProxyFactory()
            try:
                with factory.create(device_type, device_id=device_id) as proxy:
                    sp = proxy.get_service_proxy()
                    sp.quit_server()
            except ConnectionError:
                rich.print("[red]Couldn't connect to 'zonda_cs', process probably not running. ")


@app.command()
def status(device_id: str):
    """Request status information from the Control Server."""

    *_, device_type, controller_type = get_hexapod_controller_pars(device_id)

    with RegistryClient() as reg:
        service = reg.discover_service(device_id)
        # rich.print("service = ", service)

        if service:
            protocol = service.get('protocol', 'tcp')
            hostname = service['host']
            port = service['port']
            service_port = service['metadata']['service_port']
            monitoring_port = service['metadata']['monitoring_port']
            endpoint = connect_address(protocol, hostname, port)
            # rich.print(f"{endpoint = }")
        else:
            rich.print(f"[red]The ZONDA CS '{device_id}' isn't registered as a service. I cannot contact the control "
                       f"server without the required info from the service registry.[/]")
            rich.print("ZONDA Hexapod: [red]inactive")
            return

    factory = ProxyFactory()

    if is_control_server_active(endpoint):
        rich.print("ZONDA Hexapod: [green]active")
        with factory.create(device_type, device_id=device_id, protocol=protocol, hostname=hostname, port=port) as zonda:
            sim = zonda.is_simulator()
            connected = zonda.is_connected()
            ip = zonda.get_ip_address()
            rich.print(f"type: {controller_type}")
            rich.print(f"mode: {'simulator' if sim else 'device'}{'' if connected else ' not'} connected")
            rich.print(f"hostname: {ip}")
            rich.print(f"commanding port: {port}")
            rich.print(f"service port: {service_port}")
            rich.print(f"monitoring port: {monitoring_port}")

    else:
        rich.print("ZONDA Hexapod: [red]inactive")


if __name__ == "__main__":

    import logging

    from egse.logger import set_all_logger_levels
    set_all_logger_levels(logging.DEBUG)

    sys.exit(app())
